# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The wpan OpenThread CLI capability for NRF platform."""

from typing import Callable, List, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import wpan_ot_base
from gazoo_device.switchboard import expect_response

logger = gdm_logger.get_logger()

_COMMAND_RESPONSE_END = r"(.*\n)?(Done|Error \d+:.*)\n"


class WpanNrfOt(wpan_ot_base.WpanOtBase):
  """The wpan OpenThread CLI capability for NRF platform."""

  def __init__(self, device_name: str, send: Callable[..., None],
               send_and_expect: Callable[..., expect_response.ExpectResponse]):
    """Initializes an instance of WpanNrfOt capability.

    Args:
      device_name: Name of device using this capability.
      send: Switchboard.send method.
      send_and_expect: Switchboard.send_and_expect method.
    """
    super().__init__(device_name=device_name)
    self._send = send
    self._send_and_expect = send_and_expect

  @decorators.CapabilityLogDecorator(logger)
  def factory_reset(self) -> None:
    """Factory resets the Thread board.

    Raises:
      DeviceError: when something wrong happens during the factory reset.

    Note:
      factory reset command resets the board so it does not output a "DONE"
      string.
    """
    self._send(wpan_ot_base.Commands.FACTORY_RESET.value)
    self.wait_for_state({"disabled"})

  @decorators.DynamicProperty
  def csl_period(self) -> int:
    """Returns the csl period for the device in unit of 10 symbols.

    Symbol means symbol duration time in radio transmission, which equals to
    (1 / symbol rate). The symbol rate for Thread (IEEE 802.15.4, 2450MHz) is
    62.5k symbol/s. Thus 10 symbols equals to 0.00016s.
    """
    return int(self.call_command(wpan_ot_base.Commands.CSL_PERIOD.value)[0])

  @decorators.CapabilityLogDecorator(logger)
  def set_csl_period(self, csl_period: int) -> None:
    """Sets the csl period for the device in unit of 10 symbols.

    Args:
      csl_period: The csl period in unit of 10 symbols.  Symbol means symbol
        duration time in radio transmission, which equals to (1 / symbol rate).
        The symbol rate for Thread (IEEE 802.15.4, 2450MHz) is 62.5k symbol/s.
        Thus 10 symbols equals to 0.00016s.
    """
    self.call_command(f"{wpan_ot_base.Commands.CSL_PERIOD.value} {csl_period}")

  @decorators.DynamicProperty
  def csl_timeout(self) -> int:
    """Gets the csl timeout for the device.

    Returns:
      The csl timeout in unit of seconds.
    """
    return int(self.call_command(wpan_ot_base.Commands.CSL_TIMEOUT.value)[0])

  @decorators.CapabilityLogDecorator(logger)
  def set_csl_timeout(self, csl_timeout: int) -> None:
    """Sets the csl timeout for the device.

    Args:
      csl_timeout: The csl timeout in unit of seconds.
    """
    self.call_command(
        f"{wpan_ot_base.Commands.CSL_TIMEOUT.value} {csl_timeout}"
    )

  def call_command(self,
                   command: str,
                   timeout: Optional[float] = None) -> List[str]:
    """Helper method to send OTCLI command, check and return the responses."""
    if timeout is None:
      timeout = 5.0

    response = self._send_and_expect(
        command=command, pattern_list=[_COMMAND_RESPONSE_END], timeout=timeout)
    if response.timedout:
      error_message = response.before
      if "InvalidCommand" in response.before:
        raise errors.DeviceError(
            f"{self._device_name} Invalid command {command}: {error_message}")
      else:
        raise errors.DeviceError(
            f"{self._device_name} timed out responding to command {command}"
            f"after {timeout} seconds: {error_message}")
    lines = response.after.split("\n")[:-1]  # Strip the final empty line.
    if lines[-1] == "Done":
      # Command returns successfully
      return lines[1:-1]  # Strip echo back command and the status line.
    else:
      # Command returns an error
      raise errors.DeviceError(f"{self._device_name} command {command} failed. "
                               f"Response: {response.after}")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for Matter sample app controls capability.

See more details in the CHIP Linux Lighting Example document:
https://github.com/project-chip/connectedhomeip/tree/master/examples/lighting-app/linux
"""
import abc
from typing import List

from gazoo_device.capabilities.interfaces import capability_base


class MatterSampleAppBase(capability_base.CapabilityBase):
  """Abstract base class for Matter controller capability."""

  @abc.abstractmethod
  def is_present(self) -> bool:
    """Returns if the app binary is present."""

  @abc.abstractmethod
  def is_running(self) -> bool:
    """Returns if the app process is running."""

  @abc.abstractmethod
  def get_process_ids(self) -> List[str]:
    """Gets all currently active app process IDs."""

  @abc.abstractmethod
  def restart(self) -> None:
    """Restarts the app process.

    Raises:
      DeviceError if the app didn't restart successfully.
    """

  @abc.abstractmethod
  def upgrade(self, sample_app_file_path: str) -> None:
    """Upgrades to a new sample app binary.

    Args:
      sample_app_file_path: Local path to the sample app file on the host.
    """

  @abc.abstractmethod
  def factory_reset(self) -> None:
    """Factory resets all settings stored on RPi."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The bluetooth service controls capability for linux platform."""
import enum
from typing import Callable

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import bluetooth_service_base

logger = gdm_logger.get_logger()


class _Commands(str, enum.Enum):
  """System service control commands for bluetooth service."""
  STATUS = "sudo systemctl status bluetooth"
  START = "sudo systemctl start bluetooth"
  STOP = "sudo systemctl stop bluetooth"
  RESTART = "sudo systemctl restart bluetooth"


class _Regexes(str, enum.Enum):
  """Regexes for the shell output of bluetooth service."""
  STATUS = r"Active: (\w+) \(\w+\)(.*)"


class BluetoothServiceLinux(bluetooth_service_base.BluetoothServiceBase):
  """The bluetooth service controls capability for linux platform."""

  def __init__(
      self,
      device_name: str,
      shell_fn: Callable[..., str],
      shell_with_regex_fn: Callable[..., str]):
    """Initializes an instance of BluetoothServiceLinux capability.

    Args:
      device_name: Name of device using this capability.
      shell_fn: The shell method of the device.
      shell_with_regex_fn: The shell_with_regex method of the device.
    """
    super().__init__(device_name=device_name)
    self._shell = shell_fn
    self._shell_with_regex = shell_with_regex_fn

  @decorators.DynamicProperty
  def status(self) -> bool:
    """Returns whether the bluetooth service is active or not."""
    return self._shell_with_regex(
        command=_Commands.STATUS,
        regex=_Regexes.STATUS,
        raise_error=True) == "active"

  @decorators.CapabilityLogDecorator(logger)
  def start(self) -> None:
    """Starts the bluetooth service."""
    self._shell(_Commands.START)
    if not self.status:
      raise errors.DeviceError(
          f"{self._device_name} failed to start bluetooth.")

  @decorators.CapabilityLogDecorator(logger)
  def stop(self) -> None:
    """Stops the bluetooth service."""
    self._shell(_Commands.STOP)
    if self.status:  # pylint: disable=using-constant-test
      raise errors.DeviceError(
          f"{self._device_name} failed to stop bluetooth.")

  @decorators.CapabilityLogDecorator(logger)
  def restart(self) -> None:
    """Retarts the bluetooth service."""
    self._shell(_Commands.RESTART)
    if not self.status:
      raise errors.DeviceError(
          f"{self._device_name} failed to restart bluetooth.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The implementation of the package management capability on Android devices.

The Android package management capability is used on the android devices which
can have
Android APKs installed. The Android package systems rely on adb command to
manage packages.
"""
import re
from typing import Any, Callable, Dict, List, Optional
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import package_management_base
from gazoo_device.utility import adb_utils

logger = gdm_logger.get_logger()

# Dumpsys package command.
# Grep only the packages part otherwise the output size will be too large.
_COMMAND_DUMPSYS_PACKAGES = r'dumpsys package |grep "Package \[.*\]" -A 15'
_COMMAND_DUMPSYS_PACKAGE = 'dumpsys package {package_name}'
_COMMAND_LIST_PACKAGES = 'pm list packages'
_PACKAGE_PREFIX = 'package:'

# Package system info
_PACKAGE_SYSINFO_SEPARATOR_REGEX = r'\s*Package \[\S+\] \(\S+\):'
_PACKAGE_SYSINFO_HIDDEN_SYSTEM_PACKAGES_SEPARATOR = 'Hidden system packages:'
_PACKAGE_SYSINFO_VERSION_NAME_KEY = 'versionName'
_PACKAGE_SYSINFO_LAST_UPDATE_TIME_KEY = 'lastUpdateTime'
_PACKAGE_SYSINFO_RESOURCE_PATH_KEY = 'resourcePath'
_PACKAGE_SYSINFO_PACKAGE_NAME_KEY = 'pkg'
_PACKAGE_SYSINFO_PACKAGE_NAME_REGEX = r'pkg=Package{\S+ (\S+)}'
_PACKAGE_SYSINFO_VALUE_REGEX = r'{key}=(.+)'


class PackageManagementAndroid(package_management_base.PackageManagementBase):
  """Android package management capability implementation."""

  def __init__(self,
               device_name: str,
               adb_serial: str,
               shell_fn: Callable[..., Any],
               reboot_fn: Callable[..., Any]):
    """Creates an instance of the Android package management capability.

    Args:
      device_name: Name of device used for logging.
      adb_serial: The adb device serial.
      shell_fn: The function used to send shell commands to the devic.
      reboot_fn: The function to reboot the device.
    """
    super().__init__(device_name=device_name)
    self._adb_serial = adb_serial
    self._shell_fn = shell_fn
    self._reboot_fn = reboot_fn

  @decorators.CapabilityLogDecorator(logger)
  def install_package(self,
                      package_path: str,
                      allow_downgrade: bool = False,
                      allow_test_apk: bool = False,
                      reinstall: bool = False,
                      all_permissions: bool = False) -> None:
    """Installs an Android package on the device.

    Args:
      package_path: The path to the package on the host machine.
      allow_downgrade: Allows version code downgrade.
      allow_test_apk: Allows test APKs to be installed.
      reinstall: Reinstalls an existing package and keeps its data.
      all_permissions: Grants all runtime permission to the app.
    """
    adb_utils.install_package_on_device(
        package_path,
        adb_serial=self._adb_serial,
        allow_downgrade=allow_downgrade,
        allow_test_apk=allow_test_apk,
        reinstall=reinstall,
        all_permissions=all_permissions)

  @decorators.CapabilityLogDecorator(logger)
  def install_system_package(self, package_path: str, destination: str):
    """Installs an Android package as the system package on the device.

    To install a package as privileged system package, we need to remount the
    device in order to push the package to the system package folder and
    overwrite a system package. The newly added system package will take effect
    after rebooting.

    Args:
      package_path: The path to the package on the host machine.
      destination: The package destination on device.
    """
    adb_utils.remount_device(self._adb_serial, reboot_fn=self._reboot_fn)
    adb_utils.push_to_device(self._adb_serial, package_path, destination)
    self._reboot_fn()

  @decorators.CapabilityLogDecorator(logger)
  def uninstall_package(self, package_name: str) -> None:
    """Uninstalls an Android package on the device.

    Args:
      package_name: The name of the package to uninstall.
    """
    adb_utils.uninstall_package_on_device(
        package_name, adb_serial=self._adb_serial)

  @decorators.CapabilityLogDecorator(logger)
  def list_packages(self, system_package_only: bool = False) -> List[str]:
    """Lists all the installed packages on the device.

    Here we use `pm list package` command to show all the packages on the
    device. Note that the results will have be "package:<package_name>".
    We will remove the prefix here.

    Args:
      system_package_only: List only the system packages.

    Returns:
      List of package names.
    """
    cmd = _COMMAND_LIST_PACKAGES
    if system_package_only:
      cmd = f'{cmd} -s'
    cmd_result = self._shell_fn(cmd)
    packages = [
        pkg.replace(_PACKAGE_PREFIX, '') for pkg in cmd_result.split('\n')
    ]
    return packages

  @decorators.CapabilityLogDecorator(logger)
  def has_package(self, package_name: str,
                  system_package_only: bool = False) -> bool:
    """Checks if a package is installed on the device.

    Args:
      package_name: The package name.
      system_package_only: Checks if it's a system package.

    Returns:
      True if the package is installed else False.
    """
    if system_package_only:
      cmd = f'{_COMMAND_LIST_PACKAGES} -s {package_name}'
    else:
      cmd = f'{_COMMAND_LIST_PACKAGES} {package_name}'
    cmd_result = self._shell_fn(cmd)
    full_package_name = f'{_PACKAGE_PREFIX}{package_name}'
    return full_package_name in cmd_result.splitlines()

  def _get_value_from_package_sys_info(
      self,
      package_sys_info: str,
      key: str,
      regex: Optional[str] = None,
  ) -> str:
    """Gets the value from device's package system info.

    Sample package system info for a package:
    --
    Package [com.my.package] (abcd1234):
       userId=1000
       pkg=Package{abcd1234 com.my.package}
       resourcePath=/path/to/apk
       versionCode=1234 minSdk=30 targetSdk=31
       versionName=12.34.56

    Args:
      package_sys_info: The package system info from dumpsys.
      key: The value key name.
      regex: The regex for getting the value. If empty, use the general regex
        "<key>=<value>".

    Returns:
      The value.

    Raises:
      DeviceError: If failed to parse the value from the package system info.
    """
    if not regex:
      regex = _PACKAGE_SYSINFO_VALUE_REGEX.format(key=key)

    value_match = re.search(regex, package_sys_info)

    if not value_match:
      raise errors.DeviceError(
          f'{self._device_name} failed to parse {key} value from the '
          f'package system info "{package_sys_info}".')
    return value_match.group(1)

  def _get_package_system_info(
      self,
      package_name: str,
      hidden_system_package: bool = False) -> Optional[str]:
    """Gets the system info for the target package.

    Uses 'dumpsys package <package_name>' command to get package system info.
    If a package is a system package and there's another version installed, the
    pre-built package will become the hidden system package.

    Args:
      package_name: The package name.
      hidden_system_package: Return the hidden system package info.

    Returns:
      The package system info or None if the package is not installed.
        If hidden_system_package is True then returns the hidden system package
        info, or None if there is no matching hidden system package.
    """
    if not self.has_package(package_name):
      return None

    command = _COMMAND_DUMPSYS_PACKAGE.format(package_name=package_name)
    package_sys_info_group = self._shell_fn(command).split(
        _PACKAGE_SYSINFO_HIDDEN_SYSTEM_PACKAGES_SEPARATOR)

    package_sys_info = package_sys_info_group[0]
    if hidden_system_package:
      if len(package_sys_info_group) == 1:
        return None
      package_sys_info = package_sys_info_group[1]
    return package_sys_info

  @decorators.CapabilityLogDecorator(logger)
  def get_package_versions(self) -> Dict[str, str]:
    """Gets all installed package versions.

    The package version info is fetched by 'dumpsys package' command.
    The command output contains all the package related system information and
    we parse the versionName field under each package info section.

    Returns:
      A dict that contains {package_name: package_version} info.
    """
    package_sys_info = self._shell_fn(_COMMAND_DUMPSYS_PACKAGES)
    package_sys_info_by_package = re.split(
        _PACKAGE_SYSINFO_SEPARATOR_REGEX, package_sys_info)
    # Removes the first section which is before the package info list.
    package_sys_info_by_package.pop(0)
    result = {}

    for package_sys_info in package_sys_info_by_package:
      package_name = self._get_value_from_package_sys_info(
          package_sys_info=package_sys_info,
          key=_PACKAGE_SYSINFO_PACKAGE_NAME_KEY,
          regex=_PACKAGE_SYSINFO_PACKAGE_NAME_REGEX)
      version_name = self._get_value_from_package_sys_info(
          package_sys_info=package_sys_info,
          key=_PACKAGE_SYSINFO_VERSION_NAME_KEY)

      # Don't let the hidden system package version overwrite the value.
      if package_name not in result:
        result[package_name] = version_name
    return result

  @decorators.CapabilityLogDecorator(logger)
  def get_package_version(
      self,
      package_name: str,
      hidden_system_package: bool = False) -> Optional[str]:
    """Gets the version of the target package.

    Uses 'dumpsys package <package_name>' command to get the target package
    version from the package system info.
    If a package is a system package and there's another version installed, the
    pre-built package will become the hidden system package.

    Args:
      package_name: The package name.
      hidden_system_package: Return the hidden system package version.

    Returns:
      The package version or None if the package is not installed.
        If hidden_system_package=True then returns the version in the
        hidden system package info (could be None).
    """
    package_sys_info = self._get_package_system_info(
        package_name=package_name, hidden_system_package=hidden_system_package)

    if package_sys_info is None:
      return None
    return self._get_value_from_package_sys_info(
        package_sys_info=package_sys_info,
        key=_PACKAGE_SYSINFO_VERSION_NAME_KEY)

  @decorators.CapabilityLogDecorator(logger)
  def get_package_installation_time(self, package_name: str) -> Optional[str]:
    """Gets the installation time of the package.

    Args:
      package_name: The package name.

    Returns:
      The package installation time or None if the package is not installed.
        The time format is "YYYY-MM-DD HH:mm:ss" without locale info.
    """
    package_sys_info = self._get_package_system_info(package_name=package_name)

    if package_sys_info is None:
      return None
    return self._get_value_from_package_sys_info(
        package_sys_info=package_sys_info,
        key=_PACKAGE_SYSINFO_LAST_UPDATE_TIME_KEY)

  @decorators.CapabilityLogDecorator(logger)
  def get_package_resource_path(self, package_name: str) -> Optional[str]:
    """Gets the resource path of the package."""
    package_sys_info = self._get_package_system_info(package_name=package_name)

    if package_sys_info is None:
      return None
    return self._get_value_from_package_sys_info(
        package_sys_info=package_sys_info,
        key=_PACKAGE_SYSINFO_RESOURCE_PATH_KEY)

  @decorators.DynamicProperty
  def package_versions(self) -> Dict[str, str]:
    """All installed package versions."""
    return self.get_package_versions()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matter controller capability using chip-tool command.

See
https://github.com/project-chip/connectedhomeip/tree/master/examples/chip-tool
for chip-tool usage and command references.
"""

import ast
import binascii
import os
import re
import shutil
import tempfile
import time
from typing import Any, Callable, List, Optional, Sequence

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import matter_controller_base
from gazoo_device.utility import http_utils
import immutabledict
import requests

logger = gdm_logger.get_logger()

_CHIP_TOOL_BINARY_PATH = "/usr/local/bin/chip-tool"
_HEX_PREFIX = "hex:"
_MATTER_NODE_ID_PROPERTY = "matter_node_id"
_MATTER_SDK_CERTS_DIRECTORY_PATH = "https://api.github.com/repos/project-chip/connectedhomeip/contents/credentials/development/paa-root-certs/"
_MATTER_SDK_REPO = "https://api.github.com/repos/project-chip/connectedhomeip/commits/master"
_RPI_TEMP_CERTS_DIR = "/tmp/chip_certs_folder_temp"
LOGGING_FILE_PATH = "/tmp/chip.log"
DEFAULT_PAA_TRUST_STORE_PATH = "/home/pi/certs"

_COMMANDS = immutabledict.immutabledict({
    "READ_CLUSTER_ATTRIBUTE":
        "{chip_tool} {cluster} read {attribute} {node_id} {endpoint_id}",
    "WRITE_CLUSTER_ATTRIBUTE":
        "{chip_tool} {cluster} write {attribute} {value} {node_id} "
        "{endpoint_id}",
    "SUBSCRIBE_CLUSTER_ATTRIBUTE_START":
        "nohup {chip_tool} interactive start < <(echo '{cluster} subscribe "
        "{attribute} {min_interval_s} {max_interval_s} {node_id} "
        "{endpoint_id}') > {logging_file_path} 2>&1 &",
    "SUBSCRIBE_CLUSTER_ATTRIBUTE_STOP":
        "kill $(pgrep -f '{chip_tool} interactive start')",
    "READ_SUBSCRIBE_CLUSTER_ATTRIBUTE_OUTPUT":
        "grep 'Data =' {logging_file_path}",
    "SEND_CLUSTER_COMMMAND":
        "{chip_tool} {cluster} {command} {arguments} {node_id} {endpoint_id} "
        "{flags}",
    "COMMISSION_OVER_BLE_WIFI":
        "{chip_tool} pairing ble-wifi {node_id} {ssid} {password} {setup_code} "
        "{long_discriminator}",
    "COMMISSION_OVER_BLE_THREAD":
        "{chip_tool} pairing ble-thread {node_id} {operational_dataset} "
        "{setup_code} {long_discriminator}",
    "COMMISSION_ON_NETWORK":
        "{chip_tool} pairing onnetwork {node_id} {setup_code}",
    "COMMISSION_ON_NETWORK_LONG":
        "{chip_tool} pairing onnetwork-long {node_id} {setup_code} "
        "{long_discriminator}",
    "DECOMMISSION":
        "{chip_tool} pairing unpair {node_id}",
    "CHIP_TOOL_VERSION":
        "cat ~/.matter_sdk_version",
    "WRITE_CHIP_TOOL_VERSION":
        "echo {chip_tool_version} > ~/.matter_sdk_version",
    "CLEAR_TEMP_DIRECTORY":
        "rm -rf /tmp/chip*",
    "CLEAR_STORAGE":
        "{chip_tool} storage clear-all",
    "MAKE_CERTS_DIRECTORY":
        "mkdir -p {rpi_certs_dir}",
    "MOVE_CERTS_TO_CERTS_DIRECTORY":
        "mv {rpi_temp_certs_dir} {rpi_certs_dest}",
})

_REGEXES = immutabledict.immutabledict({
    "READ_CLUSTER_ATTRIBUTE_RESPONSE":
        r'Data = ([a-zA-Z0-9_+.-]+|".*")',
    "WRITE_CLUSTER_ATTRIBUTE_RESPONSE":
        r"CHIP:DMG:\s+status = (0[xX][0-9a-fA-F]+)",
    "SUBSCRIBE_CLUSTER_ATTRIBUTE_RESPONSE":
        r'Data = ([a-zA-Z0-9_+.-]+|".*")',
    "SEND_CLUSTER_COMMAND_RESPONSE":
        r"Received Command Response Status for Endpoint=\d+ "
        r"Cluster=0[xX][0-9a-fA-F_]+ Command=0[xX][0-9a-fA-F_]+ "
        r"Status=(0[xX][0-9a-fA-F]+)",
    "COMMAND_FAILURE":
        r"Run command failure: (.*)",
    "COMMISSION_SUCCESS":
        "(CHIP:TOO: Device commissioning completed with success)",
    "DECOMMISSION_COMPLETE":
        "(CHIP:DL: System Layer shutdown)",
})

_TIMEOUTS = immutabledict.immutabledict({
    "COMMISSION": 60,
    "SEND_CLUSTER_COMMAND": 30,
})


def _str_to_hex(value: str) -> str:
  """Formats a string into chip-tool compatible CLI argument.

  If the string starts with "hex:" prefix, then this function
  returns the value as is. Otherwise, the string is converted
  into its two-digit hex numbers form and prefixed with "hex:".
  For example, "password" is converted to "hex:70617373776f7264".

  See
  https://github.com/project-chip/connectedhomeip/tree/master/examples/chip-tool#using-the-client-to-commission-a-device
  for example usage.

  Args:
    value: The string to be converted.

  Returns:
    Value in its two-digit hex numbers form with "hex:" prefix.
  """
  if value.startswith(_HEX_PREFIX):
    return value

  hex_str = binascii.hexlify(str.encode(value))
  return f"{_HEX_PREFIX}{hex_str.decode()}"


def _list_certs(url: str) -> requests.Response:
  try:
    certs_list = http_utils.send_http_get(url)
  except RuntimeError as e:
    raise errors.DeviceError(
        f"Error listing files from Matter SDK({url}) "
    ) from e
  return certs_list


def _parse_response(response: str) -> Any:
  try:
    return ast.literal_eval(response)
  except ValueError:
    if response.lower() in ["true", "false"]:
      return response.lower() == "true"
    raise


class MatterControllerChipTool(matter_controller_base.MatterControllerBase):
  """Matter controller capability using chip-tool CLI commands."""

  def __init__(self,
               device_name: str,
               shell_fn: Callable[..., str],
               regex_shell_fn: Callable[..., str],
               send_file_to_device: Callable[[str, str], None],
               set_property_fn: Callable[..., None],
               get_property_fn: Callable[..., Any],
               chip_tool_path: str = _CHIP_TOOL_BINARY_PATH):
    """Creates an instance of MatterControllerChipTool capability.

    Args:
      device_name: Name of the device this capability is attached to.
      shell_fn: Bound 'shell' method of the device class instance.
      regex_shell_fn: Bound 'shell_with_regex' method of the device class
        instance.
      send_file_to_device: Bound 'send_file_to_device' method of the device's
        file transfer capability instance.
      set_property_fn: Bound 'set_property' method of the Manager instance.
      get_property_fn: Bound 'get_property' method of the Manager instance.
      chip_tool_path: Path to chip-tool binary on the device.
    """
    super().__init__(device_name)

    self._chip_tool_path = chip_tool_path
    self._shell_with_regex = regex_shell_fn
    self._shell = shell_fn
    self._send_file_to_device = send_file_to_device
    self._get_property_fn = get_property_fn
    self._set_property_fn = set_property_fn

  @decorators.DynamicProperty
  def version(self) -> str:
    """Matter SDK version of the controller."""
    return self._shell(_COMMANDS["CHIP_TOOL_VERSION"])

  @decorators.DynamicProperty
  def path(self) -> str:
    """Path to chip-tool binary."""
    return self._chip_tool_path

  @decorators.CapabilityLogDecorator(logger)
  def commission(self,
                 node_id: int,
                 setup_code: str,
                 long_discriminator: Optional[int] = None,
                 ssid: Optional[str] = None,
                 password: Optional[str] = None,
                 operational_dataset: Optional[str] = None,
                 paa_trust_store_path: Optional[str] = None) -> None:
    """Commissions a device into the controller's fabric.

    Commissioning protocol is based on specified arguments:
      - When operational dataset is provided, pairs the device over ble-thread.
      - When Wi-Fi SSID and password are provided, pairs the device over
        ble-wifi.
      - Otherwise, discover the devices on the network and pairs with the first
        one that matches the setup code and long discriminator if one is
        specified.

    Args:
      node_id: Node ID to assign to the node being commissioned.
      setup_code: Set up PIN code of the remote device.
      long_discriminator: Discriminator of the remote device.
      ssid: Wi-Fi SSID either as a string, or in the form hex:XXXXXXXX where the
        bytes of the SSID are encoded as two-digit hex numbers.
      password: Wi-Fi password, either as a string or as hex data.
      operational_dataset: Thread dataset in base-64. This argument is mutually
        exclusive with ssid and password.
      paa_trust_store_path: Path to directory holding PAA cert information
        on the controller device (e.g. rpi_matter_controller). See
        https://github.com/project-chip/connectedhomeip/tree/master/credentials/development/paa-root-certs # pylint: disable=line-too-long
        for an example.
    """
    if ssid and not password:
      raise ValueError("Wi-Fi password is not specified.")

    if operational_dataset:
      command = _COMMANDS["COMMISSION_OVER_BLE_THREAD"]
      operational_dataset = f"{_HEX_PREFIX}{operational_dataset}"
    elif ssid:
      command = _COMMANDS["COMMISSION_OVER_BLE_WIFI"]
      ssid = _str_to_hex(ssid)
      password = _str_to_hex(password)
    elif long_discriminator is not None:
      # Commission the first device found on the network with provided
      # setup code and long discriminator.
      command = _COMMANDS["COMMISSION_ON_NETWORK_LONG"]
    else:
      # Commission the first device found on the network with provided
      # setup code.
      command = _COMMANDS["COMMISSION_ON_NETWORK"]

    if paa_trust_store_path is not None:
      command += f" --paa-trust-store-path {paa_trust_store_path}"

    command = command.format(
        chip_tool=self._chip_tool_path,
        node_id=node_id,
        setup_code=setup_code,
        long_discriminator=long_discriminator,
        ssid=ssid,
        password=password,
        operational_dataset=operational_dataset,
    )
    self._shell_with_regex(
        command,
        _REGEXES["COMMISSION_SUCCESS"],
        raise_error=True,
        timeout=_TIMEOUTS["COMMISSION"])

    self._set_property_fn(_MATTER_NODE_ID_PROPERTY, node_id)

  @decorators.CapabilityLogDecorator(logger)
  def decommission(self) -> None:
    """Forgets a commissioned device with the given node id."""
    command = _COMMANDS["DECOMMISSION"].format(
        chip_tool=self._chip_tool_path,
        node_id=self._get_property_fn(_MATTER_NODE_ID_PROPERTY))
    self._shell_with_regex(
        command, _REGEXES["DECOMMISSION_COMPLETE"], raise_error=True)
    self._set_property_fn(_MATTER_NODE_ID_PROPERTY, None)

  def read(self, endpoint_id: int, cluster: str, attribute: str) -> Any:
    """Reads a cluster's attribute for the given node id and endpoint.

    Only primitive attribute values (integer, float, boolean and string)
    are supported.

    Args:
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to read the attribute value from.
      attribute: Name of the cluster attribute to read.

    Returns:
      Attribute value of the cluster.
    """
    command = _COMMANDS["READ_CLUSTER_ATTRIBUTE"].format(
        chip_tool=self._chip_tool_path,
        node_id=self._get_property_fn(_MATTER_NODE_ID_PROPERTY),
        endpoint_id=endpoint_id,
        cluster=cluster,
        attribute=attribute,
    )
    response = self._shell_with_regex(
        command, _REGEXES["READ_CLUSTER_ATTRIBUTE_RESPONSE"], raise_error=True)

    return _parse_response(response)

  @decorators.CapabilityLogDecorator(logger)
  def write(self, endpoint_id: int, cluster: str, attribute: str,
            value: Any) -> None:
    """Writes a cluster's attribute for the given node id and endpoint.

    Args:
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      value: New attribute value to update the cluster with.
    """
    command = _COMMANDS["WRITE_CLUSTER_ATTRIBUTE"].format(
        chip_tool=self._chip_tool_path,
        node_id=self._get_property_fn(_MATTER_NODE_ID_PROPERTY),
        endpoint_id=endpoint_id,
        cluster=cluster,
        attribute=attribute,
        value=value,
    )
    status_code = self._shell_with_regex(
        command, _REGEXES["WRITE_CLUSTER_ATTRIBUTE_RESPONSE"], raise_error=True)
    if int(status_code, 0) != 0:
      raise errors.DeviceError(
          f"{self._device_name} '{command}' responded with a non-zero "
          f"status code: {status_code}")

  @decorators.CapabilityLogDecorator(logger)
  def start_subscription(self, endpoint_id: int, cluster: str, attribute: str,
                         min_interval_s: int, max_interval_s: int) -> None:
    """Starts a subscription to a cluster's attribute.

    Args:
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      min_interval_s: Server should not send a new report if less than this
        number of seconds has elapsed since the last report.
      max_interval_s: Server must send a report if this number of seconds has
        elapsed since the last report.
    """
    command = _COMMANDS["SUBSCRIBE_CLUSTER_ATTRIBUTE_START"].format(
        chip_tool=self._chip_tool_path,
        node_id=self._get_property_fn(_MATTER_NODE_ID_PROPERTY),
        endpoint_id=endpoint_id,
        cluster=cluster,
        attribute=attribute,
        min_interval_s=min_interval_s,
        max_interval_s=max_interval_s,
        logging_file_path=LOGGING_FILE_PATH,
    )
    self._shell(command)

  @decorators.CapabilityLogDecorator(logger)
  def stop_subscription(self) -> List[Any]:
    """Stops existing subscription and returns subscribed values.

    Returns:
      List of attribute values.
    """
    self._shell(_COMMANDS["SUBSCRIBE_CLUSTER_ATTRIBUTE_STOP"].format(
        chip_tool=self._chip_tool_path))

    command = _COMMANDS["READ_SUBSCRIBE_CLUSTER_ATTRIBUTE_OUTPUT"].format(
        logging_file_path=LOGGING_FILE_PATH)
    responses = re.findall(_REGEXES["SUBSCRIBE_CLUSTER_ATTRIBUTE_RESPONSE"],
                           self._shell(command))
    return [_parse_response(response) for response in responses]

  @decorators.CapabilityLogDecorator(logger)
  def subscribe(self,
                endpoint_id: int,
                cluster: str,
                attribute: str,
                min_interval_s: int,
                max_interval_s: int,
                timeout_s: int = 10) -> List[Any]:
    """Subscribes to a cluster's attribute and blocks until timeout.

    Args:
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      min_interval_s: Server should not send a new report if less than this
        number of seconds has elapsed since the last report.
      max_interval_s: Server must send a report if this number of seconds has
        elapsed since the last report.
      timeout_s: Number of seconds to subscribe to this attribute for.

    Returns:
      List of attribute values.
    """
    self.start_subscription(endpoint_id, cluster, attribute, min_interval_s,
                            max_interval_s)
    time.sleep(timeout_s)
    return self.stop_subscription()

  @decorators.CapabilityLogDecorator(logger)
  def send(self,
           endpoint_id: int,
           cluster: str,
           command: str,
           arguments: Sequence[Any],
           flags: Optional[Sequence[Any]] = None) -> None:
    """Sends a command to a device with the given node id and endpoint.

    Args:
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to send the command to (e.g. onoff).
      command: Name of the command to send (e.g. toggle).
      arguments: Command arguments.
      flags: Additional flags passed to chip-tool send command. None if it's not
        provided.
    """
    additional_flags = "" if flags is None else " ".join(map(str, flags))
    command = _COMMANDS["SEND_CLUSTER_COMMMAND"].format(
        chip_tool=self._chip_tool_path,
        node_id=self._get_property_fn(_MATTER_NODE_ID_PROPERTY),
        endpoint_id=endpoint_id,
        cluster=cluster,
        command=command,
        arguments=" ".join(map(str, arguments)),
        flags=additional_flags,
    )
    command = command.rstrip()
    status_code = self._shell_with_regex(
        command,
        _REGEXES["SEND_CLUSTER_COMMAND_RESPONSE"],
        raise_error=True,
        timeout=_TIMEOUTS["SEND_CLUSTER_COMMAND"])
    if int(status_code, 0) != 0:
      raise errors.DeviceError(
          f"{self._device_name} '{command}' responded with a non-zero "
          f"status code: {status_code}")

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(self, build_file: str, build_id: str,
              paa_trust_store_path: str = DEFAULT_PAA_TRUST_STORE_PATH) -> None:
    """Installs chip-tool binary to the controller device and downloads matter certs.

    Args:
      build_file: Path to chip-tool binary on the host machine.
      build_id: Commit SHA the chip-tool binary is built at.
      paa_trust_store_path: Path where matter certs will be downloaded.
    """
    self._send_file_to_device(build_file, self._chip_tool_path)
    self._shell(
        _COMMANDS["WRITE_CHIP_TOOL_VERSION"].format(chip_tool_version=build_id))
    self.update_certs(paa_trust_store_path)

  @decorators.CapabilityLogDecorator(logger)
  def factory_reset(self) -> None:
    """Factory resets all settings stored on controller's device."""
    self._shell(_COMMANDS["CLEAR_TEMP_DIRECTORY"])
    self._shell(_COMMANDS["CLEAR_STORAGE"].format(
        chip_tool=self._chip_tool_path))

  @decorators.CapabilityLogDecorator(logger)
  def update_certs(self,
                   device_dest: str,
                   host_dest: Optional[str] = None) -> None:
    """Downloads certs from Matter SDK and to the device.

    Downloads certs and stores in the specified path in the host. Later, copies
    them to the specified path in the device.

    Args:
      device_dest: Directory path in the device to which the certs will be
        copied from the host machine.
      host_dest: Directory path in the host machine to store the certs. If not
        specified, a temporary directory will be used and cleaned up before
        exit.
    """

    with tempfile.TemporaryDirectory(prefix="chip_tool_") as tmp_dir:
      commit_sha = http_utils.send_http_get(
          _MATTER_SDK_REPO,
          headers={"Accept": "application/vnd.github.VERSION.sha"}).content
      logger.info(
          f"{self._device_name} Downloading certs from Matter SDK with commit SHA: {commit_sha.decode()}"
      )
      certs_list = _list_certs(_MATTER_SDK_CERTS_DIRECTORY_PATH).json()
      for cert in certs_list:
        response = http_utils.send_http_get(cert["download_url"])
        with open(os.path.join(tmp_dir, cert["name"]), "wb") as file:
          file.write(response.content)

      self._shell(_COMMANDS["MAKE_CERTS_DIRECTORY"].format(
          rpi_certs_dir=_RPI_TEMP_CERTS_DIR,
      ))
      self._send_file_to_device(tmp_dir, _RPI_TEMP_CERTS_DIR)
      self._shell(_COMMANDS["MAKE_CERTS_DIRECTORY"].format(
          rpi_certs_dir=device_dest,
      ))
      self._shell(_COMMANDS["MOVE_CERTS_TO_CERTS_DIRECTORY"].format(
          rpi_temp_certs_dir=os.path.join(
              _RPI_TEMP_CERTS_DIR, os.path.basename(tmp_dir), "*"),
          rpi_certs_dest=device_dest,
      ))
      if host_dest is not None:
        shutil.copytree(tmp_dir, host_dest, dirs_exist_ok=True)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the PwRPC (Pigweed RPC) wifi capability."""

from typing import Any, Callable, Dict, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_wifi_base
from gazoo_device.protos import wifi_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport


_LOGGER = gdm_logger.get_logger()
_WIFI_RPC = "WiFi"
_ASCII_NULL_BYTES = b"\x00"


class PwRPCWifiDefault(pwrpc_wifi_base.PwRPCWifiBase):
  """Pigweed RPC wifi capability."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any]):
    """Initializes an instance of the PwRPCWifiDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call

  @decorators.DynamicProperty
  def channel(self) -> int:
    """Wifi channel used with current ssid connection."""
    channel_bytes = self._wifi_rpc_call("GetChannel")
    return wifi_service_pb2.Channel.FromString(channel_bytes).channel

  @decorators.DynamicProperty
  def ssid(self) -> str:
    """Name of the SSID to which device is connected."""
    ssid_bytes = self._wifi_rpc_call("GetSsid")
    ssid = wifi_service_pb2.Ssid.FromString(ssid_bytes).ssid
    # b/212183946: Remove any null characters from ssid name.
    return ssid.rstrip(_ASCII_NULL_BYTES).decode()

  @decorators.DynamicProperty
  def state(self) -> bool:
    """Returns if the device is connected to wifi."""
    try:
      state_bytes = self._wifi_rpc_call("GetState")
      return wifi_service_pb2.State.FromString(state_bytes).connected
    except errors.DeviceError as e:
      if "Error code: 9" in str(e):
        # TODO(b/216584605) Remove error handling when the RPC is fixed.
        _LOGGER.warning(
            f"GetState doesn't work when the {self._device_name} is not "
            "connected to WiFi.")
        return False
      else:
        raise

  @decorators.DynamicProperty
  def ipv6_address(self) -> str:
    """IPv6 address for wifi interface."""
    try:
      ip6_address_bytes = self._wifi_rpc_call("GetIP6Address")
      return wifi_service_pb2.IP6Address.FromString(ip6_address_bytes).address
    except errors.DeviceError as e:
      if "Error code: 13" in str(e):
        # TODO(b/216584605) Remove error handling when the RPC is fixed.
        _LOGGER.warning(
            f"GetIP6Address doesn't work when the {self._device_name} is not "
            "connected to WiFi.")
        return ""
      else:
        raise e

  @decorators.DynamicProperty
  def mac_address(self) -> str:
    """MAC address for the wifi interface.

    Returns:
      String representation of mac_address in lower case hexadecimal digits
      separated by colon (:). e.g. 24:0a:c4:f8:6c:d4
    """
    mac_address_bytes = self._wifi_rpc_call("GetMacAddress")
    return wifi_service_pb2.MacAddress.FromString(mac_address_bytes).mac_address

  @decorators.DynamicProperty
  def wifi_interface(self) -> str:
    """Name of the interface used for wifi connection."""
    wifi_interface_bytes = self._wifi_rpc_call("GetWiFiInterface")
    return wifi_service_pb2.WiFiInterface.FromString(
        wifi_interface_bytes).interface

  @decorators.DynamicProperty
  def ipv4_address(self) -> str:
    """IPv4 address for wifi interface."""
    ip4_address_bytes = self._wifi_rpc_call("GetIP4Address")
    return wifi_service_pb2.IP4Address.FromString(ip4_address_bytes).address

  @decorators.CapabilityLogDecorator(_LOGGER)
  def connect(self,
              ssid: str,
              security_type: str,
              secret: Optional[str] = None,
              verify: bool = True) -> None:
    """Establishes wifi conection with given ssid.

    Args:
      ssid: Name of the wifi ssid.
      security_type: Wifi encryption protocol as defined in
        wifi_service_pb2.WIFI_SECURITY_TYPE.
      secret: Wifi passphrase.
      verify: Verifies if the device is connected to wifi.

    Raises:
      DeviceError with appropriate error code if connection failed.
    """
    connect_kwargs = {
        "ssid":
            ssid.encode(),
        "security_type":
            wifi_service_pb2.WIFI_SECURITY_TYPE.Value(security_type)
    }
    if secret is not None:
      connect_kwargs.update({"secret": secret.encode()})

    connection_result_bytes = self._wifi_rpc_call("Connect", connect_kwargs)
    connection_result = wifi_service_pb2.ConnectionResult.FromString(
        connection_result_bytes)
    result = connection_result.error
    if (result != wifi_service_pb2.CONNECTION_ERROR.OK or (
        verify and not self.state)):
      raise errors.DeviceError(f"{self._device_name} failed to connect to wifi "
                               f"with {result}")

  @decorators.CapabilityLogDecorator(_LOGGER)
  def disconnect(self, verify: bool = True) -> None:
    """Disconnects from current wifi connection."""
    self._wifi_rpc_call("Disconnect")
    if verify and self.state:
      raise errors.DeviceError(
          f"{self._device_name} failed to disconnect from wifi.")

  def _wifi_rpc_call(
      self,
      service_name: str,
      service_kwargs: Optional[Dict[str, Any]] = None) -> bytes:
    """Helper method for WiFi RPC call."""
    if service_kwargs is None:
      service_kwargs = {}
    return self._switchboard_call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=(_WIFI_RPC, service_name),
        method_kwargs=service_kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NRF52840 DK OpenThread device controller.

The NRF DK device acts as a full thread device which uses OpenThread CLI to
create, get and manage the thread network.
See https://openthread.io/codelabs/openthread-hardware#0 for more details.
"""

from typing import Callable, List

from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device.base_classes import nrf_connect_sdk_device
from gazoo_device.capabilities import wpan_nrf_ot
from gazoo_device.switchboard import switchboard


class NrfOpenThread(nrf_connect_sdk_device.NRFConnectSDKDevice):
  """NRF OpenThread device controller."""

  COMMUNICATION_TYPE = "JlinkSerialComms"
  _COMMUNICATION_KWARGS = {"baudrate": 115200, "enable_jlink": False}
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SerialQuery.IS_NRF_OPENTHREAD: True,
      detect_criteria.SerialQuery.PRODUCT_NAME: "j-link",
  }
  _OWNER_EMAIL = "gdm-authors@google.com"
  DEVICE_TYPE = "nrfopenthread"

  @decorators.PersistentProperty
  def health_checks(self)-> List[Callable[[], None]]:
    """Returns the list of methods to execute as health checks."""
    return super().health_checks + [self.check_otcli]

  @decorators.health_check
  def check_otcli(self) -> None:
    """Checks if the OpenThread CLI is working."""
    response = self.switchboard.send_and_expect(
        command="help", pattern_list=[".*Done"], timeout=3)
    if response.timedout:
      raise errors.DeviceBinaryMissingError(
          device_name=self.name,
          msg="The OpenThread CLI library is not working.")

  @decorators.CapabilityDecorator(switchboard.SwitchboardDefault)
  def switchboard(self) -> switchboard.SwitchboardDefault:
    """Switchboard instance for communicating with the device."""
    switchboard_name = self._get_private_capability_name(
        switchboard.SwitchboardDefault)
    if not hasattr(self, switchboard_name):
      switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
      switchboard_kwargs.update({
          "communication_address": self.communication_address,
          "communication_type": self.COMMUNICATION_TYPE,
          "log_path": self.log_file_name,
          "device_name": self.name,
          "event_parser": None})
      setattr(self, switchboard_name,
              self.get_manager().create_switchboard(**switchboard_kwargs))

    return getattr(self, switchboard_name)

  @decorators.CapabilityDecorator(wpan_nrf_ot.WpanNrfOt)
  def wpan(self) -> wpan_nrf_ot.WpanNrfOt:
    """Wpan instance for using OpenThreadCLI services."""
    return self.lazy_init(
        wpan_nrf_ot.WpanNrfOt,
        device_name=self.name,
        send=self.switchboard.send,
        send_and_expect=self.switchboard.send_and_expect)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter endpoint capability wrapper via Pigweed RPC."""
from typing import Any, Callable, List, Set, Tuple, Type

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_endpoints_and_clusters
from gazoo_device.capabilities.interfaces import matter_endpoints_base
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base
from gazoo_device.capabilities.matter_endpoints import unsupported_endpoint
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base
from gazoo_device.protos import attributes_service_pb2
from gazoo_device.protos import descriptor_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport
from gazoo_device.utility import pwrpc_utils

_DESCRIPTOR_SERVICE_NAME = "Descriptor"
_DESCRIPTOR_GET_ENDPOINTS_RPC_NAME = "PartsList"
_DESCRIPTOR_DEVICE_TYPE_RPC_NAME = "DeviceTypeList"
_DESCRIPTOR_GET_CLUSTERS_RPC_NAME = "ServerList"
_ATTRIBUTE_SERVICE_NAME = "Attributes"
_ATTRIBUTE_READ_RPC_NAME = "Read"
_ATTRIBUTE_WRITE_RPC_NAME = "Write"
_ATTRIBUTE_DATA_MODULE_PATH = "gazoo_device.protos.attributes_service_pb2.AttributeData"
_ATTRIBUTE_METADATA_MODULE_PATH = "gazoo_device.protos.attributes_service_pb2.AttributeMetadata"

logger = gdm_logger.get_logger()


class MatterEndpointsAccessorPwRpc(matter_endpoints_base.MatterEndpointsBase):
  """Capability wrapper for accessing the Matter endpoint instances."""

  _SUPPORTED_ENDPOINTS = matter_endpoints_and_clusters.SUPPORTED_ENDPOINTS

  def __init__(self, device_name: str, switchboard_call: Callable[..., Any],
               rpc_timeout_s: int):
    """Constructor of MatterEndpointsAccessorPwRpc.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method.
      rpc_timeout_s: Timeout (s) for RPC calls.
    """
    super().__init__(device_name=device_name, read=self.read, write=self.write)
    self._switchboard_call = switchboard_call
    self._rpc_timeout_s = rpc_timeout_s

  def get_supported_endpoint_ids(self) -> List[int]:
    """Gets the list of supported endpoint ids on the device."""
    list_of_supported_endpoints = self._switchboard_call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=(_DESCRIPTOR_SERVICE_NAME,
                     _DESCRIPTOR_GET_ENDPOINTS_RPC_NAME),
        method_kwargs={
            "endpoint": matter_endpoints_base.ROOT_NODE_ENDPOINT_ID,
            "pw_rpc_timeout_s": self._rpc_timeout_s})
    # TODO(b/241313435): support RootNode device type in PwRPC.
    supported_endpoint_ids = []
    for endpoint_in_bytes in list_of_supported_endpoints:
      endpoint = descriptor_service_pb2.Endpoint.FromString(endpoint_in_bytes)
      supported_endpoint_ids.append(endpoint.endpoint)
    return supported_endpoint_ids

  def get_endpoint_class_and_device_type_id(
      self, endpoint_id: int) -> Tuple[Type[endpoint_base.EndpointBase], int]:
    """Gets the endpoint class and device type ID by the given endpoint id.

    Args:
      endpoint_id: The given endpoint ID on the device.

    Returns:
      The endpoint class module (or UnsupportedEndpoint if the endpoint is not
      yet supported in GDM) and the device type ID.
    """
    device_types = self._switchboard_call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=(_DESCRIPTOR_SERVICE_NAME,
                     _DESCRIPTOR_DEVICE_TYPE_RPC_NAME),
        method_kwargs={
            "endpoint": endpoint_id, "pw_rpc_timeout_s": self._rpc_timeout_s})
    device_type = descriptor_service_pb2.DeviceType.FromString(device_types[0])
    device_type_id = device_type.device_type

    endpoint_class = (
        matter_endpoints_and_clusters.MATTER_DEVICE_TYPE_ID_TO_CLASS.
        get(device_type_id, unsupported_endpoint.UnsupportedEndpoint))

    return endpoint_class, device_type_id

  def get_supported_clusters(
      self, endpoint_id: int) -> Set[Type[cluster_base.ClusterBase]]:
    """Retrieves the supported clusters from the given endpoint ID."""
    clusters = self._switchboard_call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=(_DESCRIPTOR_SERVICE_NAME,
                     _DESCRIPTOR_GET_CLUSTERS_RPC_NAME),
        method_kwargs={
            "endpoint": endpoint_id, "pw_rpc_timeout_s": self._rpc_timeout_s})
    cluster_classes = set()
    for cluster_in_bytes in clusters:
      cluster = descriptor_service_pb2.Cluster.FromString(cluster_in_bytes)
      cluster_class = matter_endpoints_and_clusters.CLUSTER_ID_TO_CLASS_PW_RPC.get(
          cluster.cluster_id)
      if cluster_class is None:
        logger.warning(
            f"Cluster class for cluster ID {hex(cluster.cluster_id)} has not "
            "been implemented yet.")
        continue
      cluster_classes.add(cluster_class)
    return cluster_classes

  @decorators.CapabilityLogDecorator(logger)
  def read(
      self,
      endpoint_id: int,
      cluster_id: attributes_service_pb2.ClusterType,
      attribute_id: int,
      attribute_type: attributes_service_pb2.AttributeType
  ) -> attributes_service_pb2.AttributeData:
    """Ember API read method.

    Reads attribute data from the given endpoint ID, cluster ID and
    attribute ID with the given attribute type. The endpoint ID is retrieved
    via descriptor cluster. The attribute type is defined in the
    attributes_service.proto, while cluster ID and attribute ID are defined in
    the Matter spec.

    Args:
      endpoint_id: Endpoint ID to read from.
      cluster_id: Cluster ID to read from.
      attribute_id: Attribute ID to read from.
      attribute_type: Attribute data type to read.

    Returns:
      Attribute data.

    Raises:
      Device error when ack value is false.
    """
    read_kwargs = {
        "endpoint": endpoint_id, "cluster": cluster_id,
        "attribute_id": attribute_id, "type": attribute_type,
        "pw_rpc_timeout_s": self._rpc_timeout_s}

    data_in_bytes = self._switchboard_call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=(_ATTRIBUTE_SERVICE_NAME, _ATTRIBUTE_READ_RPC_NAME),
        method_kwargs=read_kwargs)

    return attributes_service_pb2.AttributeData.FromString(data_in_bytes)

  @decorators.CapabilityLogDecorator(logger)
  def write(
      self,
      endpoint_id: int,
      cluster_id: attributes_service_pb2.ClusterType,
      attribute_id: int,
      attribute_type: attributes_service_pb2.AttributeType,
      **data_kwargs: Any) -> None:
    """Ember API write method.

    Write attribute data to the given endpoint ID, cluster ID and
    attribute ID with the given attribute type. The endpoint ID is retrieved
    via descriptor cluster. The attribute type is defined in the
    attributes_service.proto, while cluster ID and attribute ID are defined in
    the Matter spec. data_kwargs is the data we want to write to the device, the
    supported data types are defined in AttributeData enum of
    attributes_service.proto.

    Args:
      endpoint_id: Endpoint ID to write to.
      cluster_id: Cluster ID to write to.
      attribute_id: Attribute ID to write to.
      attribute_type: Attribute data type to write.
      **data_kwargs: Attribute data to write.

    Raises:
      Device error when ack value is false.
    """
    data = attributes_service_pb2.AttributeData(**data_kwargs)
    metadata = attributes_service_pb2.AttributeMetadata(
        endpoint=endpoint_id,
        cluster=cluster_id,
        attribute_id=attribute_id,
        type=attribute_type)

    serialized_data = pwrpc_utils.PigweedProtoState(
        data, _ATTRIBUTE_DATA_MODULE_PATH)
    serialized_metadata = pwrpc_utils.PigweedProtoState(
        metadata, _ATTRIBUTE_METADATA_MODULE_PATH)
    write_kwargs = {"data": serialized_data, "metadata": serialized_metadata}

    self._switchboard_call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=(_ATTRIBUTE_SERVICE_NAME, _ATTRIBUTE_WRITE_RPC_NAME),
        method_kwargs=write_kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of the Matter Thermostat cluster capability.
"""
from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import thermostat_base
from gazoo_device.protos import attributes_service_pb2

logger = gdm_logger.get_logger()
ThermostatCluster = matter_enums.ThermostatCluster
ENUM8_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_ENUM8_ATTRIBUTE_TYPE
INT16S_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_INT16S_ATTRIBUTE_TYPE


class ThermostatClusterPwRpc(thermostat_base.ThermostatClusterBase):
  """Matter Thermostat cluster capability."""

  @decorators.DynamicProperty
  def local_temperature(self) -> int:
    """The LocalTemperature attribute.

    Returns:
      The temperature, as measured locally or remotely (over the network),
      on the Celsius scale with a resolution of 0.01°C.
    """
    temperature_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_LOCAL_TEMPERATURE,
        attribute_type=INT16S_ATTRIBUTE_TYPE)
    return temperature_data.data_int16

  @decorators.DynamicProperty
  def occupied_cooling_setpoint(self) -> int:
    """The OccupiedCoolingSetpoint attribute.

    Returns:
      The cooling mode setpoint when the room is occupied, on the Celsius scale
      with a resolution of 0.01°C.
    """
    setpoint_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_OCCUPIED_COOLING_SETPOINT,
        attribute_type=INT16S_ATTRIBUTE_TYPE)
    return setpoint_data.data_int16

  @occupied_cooling_setpoint.setter
  def occupied_cooling_setpoint(self, value: int) -> None:
    """Updates the OccupiedCoolingSetpoint attribute with new value."""
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_OCCUPIED_COOLING_SETPOINT,
        attribute_type=INT16S_ATTRIBUTE_TYPE,
        data_int16=value)

  @decorators.DynamicProperty
  def occupied_heating_setpoint(self) -> int:
    """The OccupiedHeatingSetpoint attribute.

    Returns:
      The heating mode setpoint when the room is occupied, on the Celsius scale
      with a resolution of 0.01°C.
    """
    setpoint_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_OCCUPIED_HEATING_SETPOINT,
        attribute_type=INT16S_ATTRIBUTE_TYPE)
    return setpoint_data.data_int16

  @occupied_heating_setpoint.setter
  def occupied_heating_setpoint(self, value: int) -> None:
    """Updates the OccupiedHeatingSetpoint attribute with new value."""
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_OCCUPIED_HEATING_SETPOINT,
        attribute_type=INT16S_ATTRIBUTE_TYPE,
        data_int16=value)

  @decorators.DynamicProperty
  def control_sequence_of_operation(
      self) -> matter_enums.ThermostatControlSequence:
    """The ControlSequenceOfOperation attribute.

    Returns:
      The overall operating environment of the thermostat, and the possible
      system modes that the thermostat can operate in.
    """
    operation_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_CONTROL_SEQUENCE_OF_OPERATION,
        attribute_type=ENUM8_ATTRIBUTE_TYPE)
    return matter_enums.ThermostatControlSequence(operation_data.data_uint8)

  @control_sequence_of_operation.setter
  def control_sequence_of_operation(
      self, value: matter_enums.ThermostatControlSequence) -> None:
    """Updates the ControlSequenceOfOperation attribute with new value."""
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_CONTROL_SEQUENCE_OF_OPERATION,
        attribute_type=ENUM8_ATTRIBUTE_TYPE,
        data_uint8=value.value)

  @decorators.DynamicProperty
  def system_mode(self) -> matter_enums.ThermostatSystemMode:
    """The SystemMode attribute.

    Returns:
      The current operating mode of the thermostat.
    """
    mode_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_SYSTEM_MODE,
        attribute_type=ENUM8_ATTRIBUTE_TYPE)
    return matter_enums.ThermostatSystemMode(mode_data.data_uint8)

  @system_mode.setter
  def system_mode(self, value: matter_enums.ThermostatSystemMode) -> None:
    """Updates the SystemMode attribute with new value."""
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=ThermostatCluster.ID,
        attribute_id=ThermostatCluster.ATTRIBUTE_SYSTEM_MODE,
        attribute_type=ENUM8_ATTRIBUTE_TYPE,
        data_uint8=value.value)

  @decorators.CapabilityLogDecorator(logger)
  def setpoint_raise_lower(self, mode: matter_enums.ThermostatSetpointMode,
                           amount: int) -> None:
    """The SetpointRaiseLower command to add specified amount to setpoint.

    As Ember API does not support cluster commands yet, this method reads the
    current occupied setpoint and updates them accordingly with setters.

    Args:
      mode: The mode for which the sepoints are to be adjusted.
      amount: The amount (possibly negative) that should be added to the
        setpoint(s), in steps of 0.1°C
    """
    delta = amount * 10
    if mode == matter_enums.ThermostatSetpointMode.HEAT:
      self.occupied_heating_setpoint = self.occupied_heating_setpoint + delta
    elif mode == matter_enums.ThermostatSetpointMode.COOL:
      self.occupied_cooling_setpoint = self.occupied_cooling_setpoint + delta
    else:
      self.occupied_heating_setpoint = self.occupied_heating_setpoint + delta
      self.occupied_cooling_setpoint = self.occupied_cooling_setpoint + delta



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Speaker endpoint.

This endpoint module corresponds to the
"Speaker" Matter device type which defines required and optional clusters
in the Matter spec 1.0.

A Speaker enpoint houses an instance of Speaker Matter device type with
the following defined clusters on this endpoint:

The required clusters for this endpoint: OnOff and Level.
"""
from typing import Union

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import level_control_chip_tool
from gazoo_device.capabilities.matter_clusters import level_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import on_off_chip_tool
from gazoo_device.capabilities.matter_clusters import on_off_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import speaker_base


class SpeakerEndpoint(speaker_base.SpeakerBase):
  """Matter Speaker endpoint."""

  @decorators.CapabilityDecorator(
      [on_off_chip_tool.OnOffClusterChipTool, on_off_pw_rpc.OnOffClusterPwRpc])
  def on_off(
      self
  ) -> Union[on_off_chip_tool.OnOffClusterChipTool,
             on_off_pw_rpc.OnOffClusterPwRpc]:
    """Matter OnOff cluster instance."""
    return self.cluster_lazy_init(matter_enums.OnOffCluster.ID)

  @decorators.CapabilityDecorator([
      level_control_chip_tool.LevelControlClusterChipTool,
      level_control_pw_rpc.LevelControlClusterPwRpc
  ])
  def level_control(
      self
  ) -> Union[level_control_chip_tool.LevelControlClusterChipTool,
             level_control_pw_rpc.LevelControlClusterPwRpc]:
    """Matter Level Control cluster instance."""
    return self.cluster_lazy_init(matter_enums.LevelControlCluster.ID)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Light Sensor endpoint.

A Light Sensor device measures and reports the pressure of a fluid.

This endpoint module corresponds to the
"Light Sensor" Matter device type which defines required and optional
clusters in the Matter spec 1.0.

A Light Sensor endpoint houses an instance of Light Sensor Matter device
type with the following defined clusters on this endpoint:

The required clusters for this endpoint: Illuminance Measurement.
"""
from typing import Union

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import illuminance_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import illuminance_measurement_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import light_sensor_base


class LightSensorEndpoint(light_sensor_base.LightSensorBase):
  """Matter Light Sensor endpoint."""

  @decorators.CapabilityDecorator([
      illuminance_measurement_chip_tool.IlluminanceMeasurementClusterChipTool,
      illuminance_measurement_pw_rpc.IlluminanceMeasurementClusterPwRpc
  ])
  def illuminance_measurement(
      self
  ) -> Union[
      illuminance_measurement_chip_tool.IlluminanceMeasurementClusterChipTool,
      illuminance_measurement_pw_rpc.IlluminanceMeasurementClusterPwRpc]:
    """Matter Illuminance Measurement cluster instance."""
    return self.cluster_lazy_init(matter_enums.IlluminanceMeasurementCluster.ID)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the Matter Basic Information cluster capability."""
import abc
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base


class BasicInformationClusterBase(
    cluster_base.ClusterBase, metaclass=abc.ABCMeta):
  """Matter Boolean State cluster capability."""

  CLUSTER_ID = matter_enums.BasicInformationCluster.ID

  @property
  @abc.abstractmethod
  def data_model_revision(self) -> int:
    """The DataModelRevision attribute."""

  @property
  @abc.abstractmethod
  def vendor_name(self) -> str:
    """The VendorName attribute."""

  @property
  @abc.abstractmethod
  def vendor_id(self) -> int:
    """The VendorID attribute."""

  @property
  @abc.abstractmethod
  def product_name(self) -> str:
    """The ProductName attribute."""

  @property
  @abc.abstractmethod
  def product_id(self) -> int:
    """The ProductID attribute."""

  @property
  @abc.abstractmethod
  def node_label(self) -> str:
    """The NodeLabel attribute."""

  @property
  @abc.abstractmethod
  def location(self) -> str:
    """The Location attribute."""

  @property
  @abc.abstractmethod
  def hardware_version(self) -> int:
    """The HardwareVersion attribute."""

  @property
  @abc.abstractmethod
  def hardware_version_string(self) -> str:
    """The HardwareVersionString attribute."""

  @property
  @abc.abstractmethod
  def software_version(self) -> int:
    """The SoftwareVersion attribute."""

  @property
  @abc.abstractmethod
  def software_version_string(self) -> str:
    """The SoftwareVersionString attribute."""

  @property
  @abc.abstractmethod
  def capability_minima(self) -> int:
    """The CapabilityMinima attribute."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The minimum base class for switchboard subprocesses."""
import contextlib
import multiprocessing
import os
import queue
import signal
import socket
import time
import traceback
import typing

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import faulthandler_utils
from gazoo_device.utility import multiprocessing_utils
from gazoo_device.utility import retry
import psutil

_PROCESS_START_TIMEOUT_S = 30
_CHILD_PROCESS_COMMAND_CONSUMPTION_TIMEOUT_S = 3
_CHILD_PROCESS_POLLING_INTERVAL_S = 0.01


def get_message(message_queue, timeout=None):
  """Returns next message from message_queue.

  Args:
      message_queue (Queue): to retrieve next message from
      timeout (float): time to wait in seconds for incoming message to arrive.

  Raises:
      ValueError: if queue provided is not a valid queue object

  Returns:
      object: The message data or None if no messages were available or
              retrieved within timeout specified.
  """
  if message_queue is None or not hasattr(message_queue, "get"):
    raise ValueError("Invalid queue of {} provided".format(type(message_queue)))

  message = None
  try:
    message = message_queue.get(block=True, timeout=timeout)
    # manager shutdown or message_queue empty
  except (IOError, queue.Empty, ValueError, socket.error):
    pass
  return message


def put_message(message_queue, message, timeout=None):
  """Puts message into message_queue using the timeout specified.

  Args:
      message_queue (Queue): to add message to
      message (object): to write into the queue
      timeout (float): time to wait in seconds for outgoing message to be
        accepted into queue.

  Raises:
      ValueError: if message_queue provided is not valid or has no put method
      Queue.Full: if outgoing message_queue is full and timeout was provided and
                  was reached before message was sent.
  """
  if message_queue is None or not hasattr(message_queue, "put"):
    raise ValueError("Invalid queue of {} provided".format(type(message_queue)))

  try:
    message_queue.put(message, block=True, timeout=timeout)
  except (IOError, socket.error):  # manager shutdown
    pass


def wait_for_queue_writes(
    message_queue: multiprocessing.Queue, timeout: float = 1) -> None:
  """Wait until the message_queue background thread writes all pending message_queue messages.

  Args:
    message_queue: Queue to wait on.
    timeout: Max seconds to wait for the message_queue buffer to become empty.

  Raises:
    RuntimeError: If the message_queue buffer did not become empty
      in the given time.
  """
  queue_buffer = message_queue._buffer  # pylint: disable=protected-access  # pytype:disable=attribute-error
  deadline = time.time() + timeout
  while time.time() < deadline and queue_buffer:
    time.sleep(0.01)
  if queue_buffer:
    raise RuntimeError(
        f"Buffer of queue {message_queue} didn't get flushed in {timeout}s.")


@contextlib.contextmanager
def _child_process_wrapper(parent_pid, process_name, device_name,
                           exception_queue):
  """Wrapper to ignore interrupts in child processes; the main process handles cleanup."""
  signal.signal(signal.SIGINT, signal.SIG_IGN)

  try:
    yield
  except EOFError:  # parent process has terminated
    pass
  except Exception:
    stack_trace = traceback.format_exc()
    try:
      exception_queue.put_nowait("Device {} raised exception in {}. "
                                 "{}".format(device_name, process_name,
                                             stack_trace))
      # Wait until the queue background thread has sent the message before
      # notifying the parent process.
      wait_for_queue_writes(exception_queue)
      os.kill(parent_pid, signal.SIGUSR1)
    except (IOError, OSError):  # queue or parent process doesn't exist anymore
      pass


def _parent_is_alive(parent_pid):
  status = False
  try:
    status = psutil.Process(parent_pid).status() != psutil.STATUS_ZOMBIE
  except psutil.NoSuchProcess:  # parent process doesn't exist anymore
    pass
  return status


def _process_loop(cls, parent_pid):
  """Child process loop which handles start/stop events and exceptions."""
  with _child_process_wrapper(parent_pid, cls.process_name, cls.device_name,
                              cls._exception_queue):
    faulthandler_utils.set_up_faulthandler(
        typing.cast(str, cls._log_directory),  # pylint: disable=protected-access
        log_file_name_prefix=f"{cls.device_name}_{type(cls).__name__}_parent_"
                             f"{parent_pid}_child")
    gdm_logger.initialize_child_process_logging(cls.logging_queue)
    try:
      cls._start_event.set()
    except Exception as err:
      stack_trace = traceback.format_exc()
      gdm_logger.get_logger().info(
          "Device {} Process {} error {!r} start event error. {}".format(
              cls.device_name, cls.process_name, err, stack_trace))
    running = cls._pre_run_hook()
    while running and _parent_is_alive(parent_pid):
      try:
        if cls._terminate_event.is_set():
          cls._terminate_event.clear()
          break
      except IOError:  # manager shutdown
        break
      running = cls._do_work()
  try:
    cls._stop_event.set()
  except IOError:  # manager shutdown
    pass
  cls._post_run_hook()


class SwitchboardProcess:
  """Simplifies creating Switchboard processes.

  Provides the following capabilities:
      * starting and verifying process start
      * stopping and verifying process stop
      * checking if process was started
      * checking if process is still running
      * providing pre and post process loop hooks
      * signaling main process when exceptions are raised in subprocess
      * receiving queue commands with optional data to be processed
      * Simplifying process loop definition for subclasses
  """

  def __init__(self,
               device_name,
               process_name,
               exception_queue,
               command_queue,
               log_path,
               valid_commands=None):
    """Initialize SwitchboardProcess with the arguments provided.

    Args:
        device_name (str): of the device for exception error messages.
        process_name (str): to use for process name and exception error
          messages.
        exception_queue (Queue): to use for reporting exception traceback
          message from subprocess.
        command_queue (Queue): to receive commands into.
        log_path (str): path and filename to write log messages to.
        valid_commands (Optional[Tuple[str, ...]]): Valid command strings.
    """
    log_directory = os.path.dirname(log_path)
    self.device_name = device_name
    self.process_name = process_name
    self._command_queue = command_queue
    self._exception_queue = exception_queue
    self._log_directory = log_directory
    gdm_logger.switch_to_multiprocess_logging()
    self.logging_queue = gdm_logger.get_logging_queue()
    self._start_event = multiprocessing_utils.get_context().Event()
    self._stop_event = multiprocessing_utils.get_context().Event()
    self._terminate_event = multiprocessing_utils.get_context().Event()
    self._valid_commands = valid_commands or ()
    self._process = None

  def __del__(self):
    if self.is_started():
      self.stop()

  def start(self, wait_for_start: bool = True) -> None:
    """Starts the process.

    Args:
        wait_for_start: Whether to wait for the process to start. If False, the
            caller is responsible for calling wait_for_start() separately.

    Raises:
        RuntimeError: if the process has already been started or fails to start.
    """
    if not self.is_started():
      self._start_event.clear()
      self._stop_event.clear()
      parent_pid = os.getpid()
      process = multiprocessing_utils.get_context().Process(  # pytype: disable=attribute-error  # re-none
          name=self.process_name,
          target=_process_loop,
          args=(self, parent_pid))
      process.start()
      if wait_for_start:
        self.wait_for_start()
      self._process = process
    else:
      raise RuntimeError("Device {} failed to start child process {}. "
                         "Child process is already running.".format(
                             self.device_name, self.process_name))

  def wait_for_start(self) -> None:
    """Waits for the process to start.

    Raises:
      RuntimeError: process failed to start before a timeout was reached.
    """
    start_event_value = self._start_event.wait(timeout=_PROCESS_START_TIMEOUT_S)
    if not start_event_value:
      per_core_cpu_utilization = psutil.cpu_percent(interval=0.1, percpu=True)
      per_core_cpu_utilization_formatted = [
          f"{value:.1f}%" for value in per_core_cpu_utilization]
      num_cores = len(per_core_cpu_utilization)
      avg_cpu_utilization = sum(per_core_cpu_utilization) / num_cores
      # Average load for the last 1 minute, 5 minutes, 15 minutes.
      # The "load" is defined as the average number of processes in a runnable
      # state per CPU core. Load over 100% indicates throttling.
      avg_historic_load = [f"{load / num_cores * 100:.1f}%"
                           for load in psutil.getloadavg()]
      raise RuntimeError(
          f"{self.device_name} failed to start child process "
          f"{self.process_name}. "
          f"Start event was not set in {_PROCESS_START_TIMEOUT_S}s. "
          "This typically means that the test binary has not been allocated a "
          "sufficient share of the host's CPU. Common causes: too many "
          "simultaneous task executions or process leaks from previous tests.\n"
          "Check the host CPU utilization and active processes by running "
          "'top' on the host. Also check the CPU model and frequency via "
          "'lscpu'.\n"
          "Diagnostic information: "
          f"host CPU cores: {num_cores}, "
          f"current combined CPU load: {avg_cpu_utilization:.1f}, "
          f"current per-core CPU load: {per_core_cpu_utilization_formatted}, "
          "average combined CPU load in the last 1 minute, 5 minutes, 15 "
          f"minutes: {avg_historic_load}. Load over 100% indicates throttling.")

  def is_started(self):
    """Returns True if process was started, False otherwise.

    Return:
        bool: True if process was started
    """
    return self._process is not None

  def is_running(self):
    """Returns True if process is currently running, False otherwise.

    Return:
        bool: True if process was started and is currently running

    Note:
        Use this method to determine if a process is still running, since
        the process might choose to self exit.
    """
    if self.is_started():
      return self._process.is_alive()
    return False

  def send_command(self, command, data=None,
                   wait_for_command_consumption: bool = False):
    """Sends command with optional data provided.

    Args:
        command (str): string to send
        data (object): optional data to include with command.
        wait_for_command_consumption: wait until command is consummed.

    Raises:
        ValueError: if command provided is not in the list of valid commands
    """
    if command not in self._valid_commands:
      raise ValueError("Device {} send command to child process {} failed. "
                       "{} is not a valid command in {!r}".format(
                           self.device_name, self.process_name, command,
                           self._valid_commands))
    put_message(self._command_queue, (command, data))
    if wait_for_command_consumption:
      try:
        retry.retry(
            func=self.is_command_consumed,
            is_successful=retry.is_true,
            timeout=_CHILD_PROCESS_COMMAND_CONSUMPTION_TIMEOUT_S,
            interval=_CHILD_PROCESS_POLLING_INTERVAL_S)
      except errors.CommunicationTimeoutError as err:
        self.stop()
        raise errors.ProcessCommunicationError(
            self.device_name,
            "Device {} Process {} command {} consumption timed out. "
            "Stopped the process.".format(
                self.device_name, self.process_name, command)) from err

  def stop(self):
    """Stops process if process is running.

    Raises:
        RuntimeError: if called when process is not currently running
        OSError: if a system related exception occurs.

    Note:
        When calling this method you should use is_started() to check if
        process was previously started to prevent raising an error.
    """
    if self.is_started():
      if self._process.is_alive():
        try:
          self._terminate_event.set()
        except IOError:  # manager shutdown
          pass
      try:
        stop_event_value = self._stop_event.wait(timeout=5)
        if not stop_event_value:
          msg = ("Device {} failed to stop child process {}. "
                 "Stop event was not set.").format(self.device_name,
                                                   self.process_name)
          gdm_logger.get_logger().error(msg)
          raise IOError(msg)
      except (IOError, ValueError):  # manager shutdown failed
        pass
      self._process.join(timeout=1)
      if self._process.is_alive():
        self.terminate()
      else:
        self._process = None
    else:
      msg = ("Device {} failed to stop child process {}. Child process is not "
             "currently running.").format(self.device_name, self.process_name)
      gdm_logger.get_logger().error(msg)
      raise RuntimeError(msg)

  def terminate(self):
    """Terminates the process."""
    self._process.terminate()
    self._process.join(timeout=1)
    self._process = None

  def is_command_consumed(self):
    """Returns True if command queue is empty, false otherwise.

    This does not guarantee that the command processing is complete, only that
    it has started.

    Returns:
      bool: True if command queue is empty.
    """
    return self._command_queue.empty()

  def _do_work(self):
    """Perform subprocess work and indicate if process should continue.

    Returns:
        bool: indicating process should continue running

    Note:
        This implementation just returns False.
    """
    return False

  def _post_run_hook(self):
    """Cleanup hook for derived classes after process terminates.

    Note:
        Try to avoid raising any exceptions from this method as any exception
        raised will not reach the main process.
    """
    pass

  def _pre_run_hook(self):
    """Setup hook for derived classes before child process run loop begins.

    Returns:
        bool: A boolean value indicating the process loop should proceed.

    Note:
        Returning None or False from this method will prevent the process loop
        from proceeding.
    """
    return True



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Thermostat endpoint.

This endpoint module corresponds to the
"Thermostat" Matter device type which defines required and optional clusters
in the Matter spec 1.0.

A Thermostat endpoint houses an instance of Thermostat Matter device type with
the following defined clusters on this endpoint:

The required clusters for this endpoint: Thermostat.
The optional clusters for this endpoint: Groups, Scenes, Alarms,
Thermostat User Interface Configuration, Relative Humidity Measurement,
Fan Control, Time, TimeSync, Temperature Measurement, Occupancy Sensing.
"""
from typing import Union

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import fan_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import occupancy_sensing_chip_tool
from gazoo_device.capabilities.matter_clusters import occupancy_sensing_pw_rpc
from gazoo_device.capabilities.matter_clusters import relative_humidity_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import temperature_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import thermostat_chip_tool
from gazoo_device.capabilities.matter_clusters import thermostat_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import thermostat_base


class ThermostatEndpoint(thermostat_base.ThermostatBase):
  """Matter Thermostat endpoint."""

  @decorators.CapabilityDecorator([
      thermostat_chip_tool.ThermostatClusterChipTool,
      thermostat_pw_rpc.ThermostatClusterPwRpc
  ])
  def thermostat(
      self
  ) -> Union[thermostat_chip_tool.ThermostatClusterChipTool,
             thermostat_pw_rpc.ThermostatClusterPwRpc]:
    """Matter Thermostat cluster instance."""
    return self.cluster_lazy_init(matter_enums.ThermostatCluster.ID)

  @decorators.CapabilityDecorator(
      relative_humidity_measurement_pw_rpc.
      RelativeHumidityMeasurementClusterPwRpc)
  def relative_humidity_measurement(
      self) -> (relative_humidity_measurement_pw_rpc.
                RelativeHumidityMeasurementClusterPwRpc):
    """Matter Relative Humidity Measurement cluster instance."""
    return self.cluster_lazy_init(
        matter_enums.RelativeHumidityMeasurementCluster.ID)

  @decorators.CapabilityDecorator(
      temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc)
  def temperature_measurement(
      self
  ) -> temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc:
    """Matter Temperature Measurement cluster instance."""
    return self.cluster_lazy_init(matter_enums.TemperatureMeasurementCluster.ID)

  @decorators.CapabilityDecorator([
      occupancy_sensing_chip_tool.OccupancySensingClusterChipTool,
      occupancy_sensing_pw_rpc.OccupancySensingClusterPwRpc
  ])
  def occupancy_sensing(
      self
  ) -> Union[occupancy_sensing_chip_tool.OccupancySensingClusterChipTool,
             occupancy_sensing_pw_rpc.OccupancySensingClusterPwRpc]:
    """Matter Occupancy Sensing cluster instance."""
    return self.cluster_lazy_init(matter_enums.OccupancySensingCluster.ID)

  @decorators.CapabilityDecorator(fan_control_pw_rpc.FanControlClusterPwRpc)
  def fan_control(self) -> fan_control_pw_rpc.FanControlClusterPwRpc:
    """Matter Fan Control cluster instance."""
    return self.cluster_lazy_init(matter_enums.FanControlCluster.ID)

  # TODO(b/239741839): add optional clusters below
  # def scenes(self):
  #   """Matter scenes cluster instance."""

  # def groups(self):
  #   """Matter groups cluster instance."""

  # def alarms(self):
  #   """Matter alarms cluster instance."""

  # def thermostat_user_interface_configuration(self):
  #   """Matter thermostat UI configuration cluster instance."""

  # def time(self):
  #   """Matter time cluster instance."""

  # def time_sync(self):
  #   """Matter time sync cluster instance."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device controller for RPi Matter device.

This device controller populates the supported Matter endpoints on the RPi
platform by using the descriptor RPC service.
"""
import time

from typing import Callable, Dict, List, Optional, Tuple
from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import matter_device_base
from gazoo_device.base_classes import ssh_device
from gazoo_device.capabilities import bluetooth_service_linux
from gazoo_device.capabilities import matter_sample_app_shell
from gazoo_device.capabilities.interfaces import matter_controller_base
from gazoo_device.protos import attributes_service_pb2
from gazoo_device.protos import descriptor_service_pb2
from gazoo_device.protos import device_service_pb2
from gazoo_device.utility import host_utils
from gazoo_device.utility import pwrpc_utils
from gazoo_device.utility import retry
import immutabledict

logger = gdm_logger.get_logger()

_APP_START_SEC = 1  # seconds
_POLL_APP_INTERVAL_SEC = 0.1  # seconds
_RPC_START_SEC = 5  # seconds
_POLL_RPC_INTERVAL_SEC = 1  # seconds

COMMANDS = immutabledict.immutabledict({
    "REBOOT": "sudo reboot",
    "SERIAL_NUMBER_INFO": "cat /proc/cpuinfo",
})

REGEXES = immutabledict.immutabledict({
    "SERIAL_NUMBER_INFO_REGEX": r"Serial\s+: ([^\n]+)"
})


class RaspberryPiMatter(
    ssh_device.SshDevice, matter_device_base.MatterDeviceBase):
  """RPi Matter device controller."""
  COMMUNICATION_TYPE = "PigweedSocketComms"
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SshQuery.IS_UBUNTU_RPI: True,
      detect_criteria.SshQuery.IS_MATTER_LINUX_APP_RUNNING: True,
  }
  _COMMUNICATION_KWARGS = {"protobufs": (attributes_service_pb2,
                                         descriptor_service_pb2,
                                         device_service_pb2),
                           "port": pwrpc_utils.MATTER_LINUX_APP_DEFAULT_PORT,
                           "args": host_utils.DEFAULT_SSH_OPTIONS,
                           "log_cmd": ssh_device.COMMANDS["LOGGING"],
                           "key_info": config.KEYS["raspberrypi3_ssh_key"],
                           "username": "ubuntu"}
  DEVICE_TYPE = "rpimatter"
  # Overrides to recover from the successive recoverable health check failures
  _RECOVERY_ATTEMPTS = 3

  # RPi Matter controller has transport number 2 for RPC interaction.
  _PIGWEED_PORT = 2

  MATTER_COMMISSION_METHOD = matter_controller_base.CommissionMethod.ON_NETWORK

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._commands.update(COMMANDS)
    self._regexes.update(REGEXES)

  @decorators.PersistentProperty
  def os(self) -> str:
    """OS of RPi linux app."""
    return "Ubuntu"

  @decorators.PersistentProperty
  def platform(self) -> str:
    """Platform of RPi linux app."""
    return "Raspberry Pi 4"

  @decorators.LogDecorator(logger)
  def get_detection_info(self) -> Tuple[Dict[str, str], Dict[str, str]]:
    """Gets the persistent and optional attributes of a device during setup.

    Returns:
      Dictionary of persistent attributes and dictionary of optional attributes.
    """
    persistent_dict, options_dict = super().get_detection_info()
    serial_number = self.shell_with_regex(
        self.commands["SERIAL_NUMBER_INFO"],
        self.regexes["SERIAL_NUMBER_INFO_REGEX"])
    persistent_dict["serial_number"] = serial_number
    persistent_dict["model"] = "PROTO"
    return persistent_dict, options_dict

  @decorators.PersistentProperty
  def health_checks(self) -> List[Callable[[], None]]:
    """Returns the list of methods to execute as health checks."""
    return super().health_checks + [
        self.check_app_present,
        self.check_has_service,
        self.check_is_service_enabled,
        self.check_app_running,
        self.check_open_pwrpc_socket_transport,
        self.check_rpc_working]

  @decorators.health_check
  def check_app_present(self) -> None:
    """Checks if the Matter sample app is present."""
    if not self.matter_sample_app.is_present:
      raise errors.DeviceBinaryMissingError(
          device_name=self.name,
          msg="The Matter sample app binary does not exist.")

  @decorators.health_check
  def check_has_service(self) -> None:
    """Checks if the Matter sample app service file exists."""
    if not self.matter_sample_app.has_service:
      raise errors.DeviceMissingPackagesError(
          device_name=self.name,
          msg="The Matter sample app service file does not exist.",
          package_list=(
              f"/etc/systemd/system/{pwrpc_utils.MATTER_LINUX_APP_NAME}"
              ".service",))

  @decorators.health_check
  def check_is_service_enabled(self) -> None:
    """Checks if the Matter sample app service is enabled."""
    if not self.matter_sample_app.is_service_enabled:
      raise errors.ServiceNotEnabledError(
          device_name=self.name,
          msg="The Matter sample app service is not enabled.")

  @decorators.health_check
  def check_app_running(self) -> None:
    """Checks if the Matter sample app is running."""
    if not self.matter_sample_app.is_running:
      raise errors.ProcessNotRunningError(
          device_name=self.name,
          msg="The Matter sample app process is not running.")

  @decorators.health_check
  def check_open_pwrpc_socket_transport(self) -> None:
    """Checks if the PwRpc socket transport can be opened successfully.

    Note that this health check can only be run after the health checks which
    ensure the sample app is running (check_app_present and check_app_running).
    """
    self.switchboard.open_transport(port=self._PIGWEED_PORT)

  @decorators.CapabilityDecorator(matter_sample_app_shell.MatterSampleAppShell)
  def matter_sample_app(self):
    """Matter sample app control and management capability over shell."""
    return self.lazy_init(
        matter_sample_app_shell.MatterSampleAppShell,
        device_name=self.name,
        shell_fn=self.shell,
        close_transport_fn=self.switchboard.close_transport,
        open_transport_fn=self.switchboard.open_transport,
        pigweed_port=self._PIGWEED_PORT,
        send_file_to_device_fn=self.file_transfer.send_file_to_device,
        wait_for_bootup_complete_fn=self.wait_for_bootup_complete,
        reset_endpoints_fn=self.matter_endpoints.reset)

  @decorators.LogDecorator(logger)
  def recover(self, error: errors.DeviceError) -> None:
    """Recovers the device from an error.

    1. Enables the sample app service when it's not enabled.
    2. Starts the sample app service if the process is not running.
    3. Otherwise trying with a parent recover method.

    Args:
      error: The error we try to recover from.
    """
    if isinstance(error, errors.ServiceNotEnabledError):
      self.matter_sample_app.enable_service()
    elif isinstance(error, errors.ProcessNotRunningError):
      self.shell(matter_sample_app_shell.COMMANDS["START_SERVICE"])
      # Wait a bit for the process to start.
      time.sleep(_APP_START_SEC)
    else:
      super().recover(error)

  @decorators.LogDecorator(logger)
  def reboot(self, no_wait: bool = False, method: str = "shell") -> None:
    """Reboots the RPi.

    Args:
      no_wait: Return before reboot completes.
      method: Not used.
    """
    del method
    self._inject_log_marker()
    self.switchboard.send(command=self.commands["REBOOT"])
    if not no_wait:
      self._verify_reboot()

  @decorators.LogDecorator(logger)
  def factory_reset(self, no_wait: bool = False) -> None:
    """Factory resets the device."""
    self.matter_sample_app.factory_reset()
    super().factory_reset(no_wait)

  @decorators.LogDecorator(logger)
  def wait_for_bootup_complete(self, timeout: Optional[int] = None) -> None:
    """Waits until the device finished bootup and becomes responsive."""
    # Wait for the RPi to be back online.
    super().wait_for_bootup_complete(timeout)

    # Wait for the sample app process to run again.
    retry.retry(
        func=self.check_app_running,
        timeout=_APP_START_SEC,
        interval=_POLL_APP_INTERVAL_SEC,
        reraise=False)

    # Wait for the sample app to become RPC responsive
    retry.retry(
        func=self.check_rpc_working,
        timeout=_RPC_START_SEC,
        interval=_POLL_RPC_INTERVAL_SEC,
        reraise=False)

  @decorators.CapabilityDecorator(bluetooth_service_linux.BluetoothServiceLinux)
  def bluetooth_service(self):
    """Bluetooth service controls capability."""
    return self.lazy_init(
        bluetooth_service_linux.BluetoothServiceLinux,
        device_name=self.name,
        shell_fn=self.shell,
        shell_with_regex_fn=self.shell_with_regex)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for using the faulthandler.

This module installs fault handlers which dump tracebacks for all active threads
periodically and whenever SIGTERM is received. This is useful for debugging
unexpected process terminations and hanging processes.
"""
# import atexit
# import datetime
import faulthandler
import io
import os

from gazoo_device.utility import signal_utils

_FAULTHANDLER_PERIOD_S = 60 * 5


def _stop_traceback_dumps(file: io.TextIOWrapper) -> None:
  """Stops traceback dumps and removes file if empty or sigterm hasn't occurred.

  This function will run at the end of the process only if there is no unhandled
  signal. Since sigerm is handled by GDM, it will also run on receiving sigterm
  signal. For more details, refer to b/262596084.
  - We want to preserve faulthandler logs in the event we have a hung process
  that is killed by a sigterm signal.
  - We want to remove faulthandler logs if the process exited gracefully since
  there is no use for the faulthandler logs of a process that exited
  gracefully.

  Args:
    file: File containing the faulthandlers logs.
  """
  faulthandler.cancel_dump_traceback_later()
  file.close()
  if not os.stat(file.name).st_size or not signal_utils.has_sigterm_occurred():
    os.remove(file.name)


def set_up_faulthandler(folder: str,
                        log_file_name_prefix: str = "",
                        repeat_timeout_s: int = _FAULTHANDLER_PERIOD_S) -> None:
  """Sets up faulthandler for logging."""
  # TODO(b/268778945): Disable faulthandler to investigate
  del folder
  del log_file_name_prefix
  del repeat_timeout_s
  return
  # Setup periodic traceback dumping to investigate hanging issues.
  # pid = os.getpid()
  # time_string = datetime.datetime.now().isoformat().replace(":", "-")
  # os.makedirs(folder, exist_ok=True)
  # prefix = "periodic_stack_trace_dumps"
  # if log_file_name_prefix:
  #   prefix += "_" + log_file_name_prefix
  # file_handle = open(
  #     os.path.join(folder, f"{prefix}_{pid}_{time_string}.log"), "w")
  # faulthandler.enable(file=file_handle)
  # faulthandler.dump_traceback_later(
  #     repeat=True, timeout=repeat_timeout_s, file=file_handle)
  # atexit.register(_stop_traceback_dumps, file_handle)



"""chip-tool implementation of the Matter endpoints accessor capability.

MatterEndpointsAccessorChipTool discovers the available endpoints on a Matter
device and exposes corresponding cluster capabilities, based on the results
retrieved via RaspberryPiMatterController using chip-tool binary. Communications
with the Matter end device are done via Matter protocol, and the end device is
expected to be commissioned by the same RaspberryPiMatterController. Since
chip-tool only supports commissioning a single device at a time, this capability
is attached to RaspberryPiMatterController instead of the Matter end device
for simplicity.
"""

import re
from typing import Callable, List, Set, Tuple, Type

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_endpoints_and_clusters
from gazoo_device.capabilities.interfaces import matter_controller_base
from gazoo_device.capabilities.interfaces import matter_endpoints_base
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base
from gazoo_device.capabilities.matter_endpoints import unsupported_endpoint
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base
import immutabledict

logger = gdm_logger.get_logger()

_COMMANDS = immutabledict.immutabledict({
    "READ_DESCRIPTOR_PARTS_LIST":
        "{chip_tool} descriptor read parts-list {node_id} {endpoint_id}",
    "READ_DESCRIPTOR_SERVER_LIST":
        "{chip_tool} descriptor read server-list {node_id} {endpoint_id}",
    "READ_DESCRIPTOR_DEVICE_LIST":
        "{chip_tool} descriptor read device-list {node_id} {endpoint_id}",
    "READ_DESCRIPTOR_DEVICE_TYPE_LIST":
        "{chip_tool} descriptor read device-type-list {node_id} {endpoint_id}",
})

_REGEXES = immutabledict.immutabledict({
    "DESCRIPTOR_ATTRIBUTE_RESPONSE": r"CHIP:DMG:\s+Data = (\w+)",
    "DEVICE_LIST_RESPONSE": r"CHIP:TOO:\s+(?:Device)?Type: (\d+)",
})


class MatterEndpointsAccessorChipTool(matter_endpoints_base.MatterEndpointsBase
                                     ):
  """Capability for accessing the Matter endpoint instances via chip-tool."""

  _SUPPORTED_ENDPOINTS = matter_endpoints_and_clusters.SUPPORTED_ENDPOINTS

  def __init__(
      self,
      device_name: str,
      node_id_getter: Callable[[], int],
      shell_fn: Callable[..., str],
      shell_with_regex: Callable[..., str],
      matter_controller: matter_controller_base.MatterControllerBase,
  ) -> None:
    """Initializes an instance of MatterEndpoints capability.

    Args:
      device_name: Name of the device instance the capability is attached to.
      node_id_getter: Getter method for Matter node ID of the commissioned
        end device.
      shell_fn: Bound 'shell' method of the device class instance.
      shell_with_regex: Bound 'shell_with_regex' method of the device class
        instance.
      matter_controller: An instance of MatterController capability.
    """
    super().__init__(
        device_name=device_name,
        read=matter_controller.read,
        write=matter_controller.write,
        send=matter_controller.send)

    self._matter_controller = matter_controller
    self._node_id_getter = node_id_getter
    self._shell_fn = shell_fn
    self._shell_with_regex = shell_with_regex

  def get_supported_endpoint_ids(self) -> List[int]:
    """Returns the list of supported endpoint ids on the device."""
    response = self._shell_fn(_COMMANDS["READ_DESCRIPTOR_PARTS_LIST"].format(
        chip_tool=self._matter_controller.path,
        endpoint_id=matter_endpoints_base.ROOT_NODE_ENDPOINT_ID,
        node_id=self._node_id_getter()))
    endpoints = re.findall(_REGEXES["DESCRIPTOR_ATTRIBUTE_RESPONSE"], response)
    # Descriptor cluster does not explicitly list root node endpoint.
    endpoint_ids = [matter_endpoints_base.ROOT_NODE_ENDPOINT_ID]
    endpoint_ids += [int(endpoint) for endpoint in endpoints]
    return endpoint_ids

  def get_endpoint_class_and_device_type_id(
      self, endpoint_id: int) -> Tuple[Type[endpoint_base.EndpointBase], int]:
    """Gets the endpoint class and device type ID by the given endpoint id.

    Args:
      endpoint_id: The given endpoint ID on the device.

    Returns:
      The endpoint class module (or UnsupportedEndpoint if the endpoint is not
      yet supported in GDM) and the device type ID.

    Raises:
      DeviceError when the device type ID of this endpoint cannot be obtained.
    """
    device_type_id = None
    # Different command syntax between 1.0 and master branches.
    for device_type_command in (
        "READ_DESCRIPTOR_DEVICE_LIST", "READ_DESCRIPTOR_DEVICE_TYPE_LIST"):
      command = _COMMANDS[device_type_command].format(
          chip_tool=self._matter_controller.path,
          endpoint_id=endpoint_id,
          node_id=self._node_id_getter())
      output = self._shell_with_regex(command, _REGEXES["DEVICE_LIST_RESPONSE"])
      if output.isdigit():
        device_type_id = int(output)
        break
    if device_type_id is None:
      raise errors.DeviceError(
          f"Failed to get device type from endpoint {endpoint_id} on "
          f"{self._device_name}")

    endpoint_class = (
        matter_endpoints_and_clusters.MATTER_DEVICE_TYPE_ID_TO_CLASS.
        get(device_type_id, unsupported_endpoint.UnsupportedEndpoint))

    return endpoint_class, device_type_id

  def get_supported_clusters(
      self, endpoint_id: int) -> Set[Type[cluster_base.ClusterBase]]:
    """Retrieves the supported clusters from the given endpoint ID.

    Args:
      endpoint_id: The given endpoint ID on the device.

    Returns:
      Set of supported cluster capability classes.
    """
    response = self._shell_fn(_COMMANDS["READ_DESCRIPTOR_SERVER_LIST"].format(
        chip_tool=self._matter_controller.path,
        endpoint_id=endpoint_id,
        node_id=self._node_id_getter()))
    clusters = map(
        int, re.findall(_REGEXES["DESCRIPTOR_ATTRIBUTE_RESPONSE"], response))

    cluster_classes = []
    for cluster in clusters:
      if cluster in matter_endpoints_and_clusters.CLUSTER_ID_TO_CLASS_CHIP_TOOL:
        cluster_classes.append(matter_endpoints_and_clusters
                               .CLUSTER_ID_TO_CLASS_CHIP_TOOL[cluster])
      else:
        logger.warning(f"Cluster class for cluster ID {hex(cluster)} has not "
                       "been implemented yet.")

    return set(cluster_classes)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the Matter cluster capability."""
import inspect
from typing import Any, Callable, Optional
from gazoo_device.capabilities.interfaces import capability_base


class ClusterBase(capability_base.CapabilityBase):
  """Matter cluster capability interface."""

  # Cluster ID defined in the Matter spec.
  CLUSTER_ID = None

  def __init__(self,
               device_name: str,
               endpoint_id: int,
               read: Callable[..., Any],
               write: Callable[..., Any],
               send: Optional[Callable[..., Any]] = None):
    """Initializes an instance of the Matter cluster capability.

    Args:
      device_name: Device name used for logging.
      endpoint_id: Endpoint ID of endpoint which hosts this cluster.
      read: The Ember API or MatterController read method.
      write: The Ember API or MatterController write method.
      send: The MatterController command send method.
    """
    super().__init__(device_name=device_name)
    self._endpoint_id = endpoint_id
    self._read = read
    self._write = write
    self._send = send

  def __setattr__(self, name: str, value: Any) -> None:
    """Overrides the __setattr__ to check if setting to the valid attribute."""
    def is_setter_property(member):
      return isinstance(member, property) and member.fset is not None

    valid_attributes = {
        name for name, _ in inspect.getmembers(
            self.__class__, is_setter_property)}

    # Skip the private attribute: _device_name, _healthy, _endpoint_id etc.
    if name not in valid_attributes and not name.startswith("_"):
      raise AttributeError(
          f"Invalid attribute '{name}' to set. Valid attributes are "
          f"{valid_attributes}.")

    super().__setattr__(name, value)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the Matter Thermostat cluster capability."""
import abc
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base


class ThermostatClusterBase(cluster_base.ClusterBase, metaclass=abc.ABCMeta):
  """Matter Thermostat cluster capability."""

  CLUSTER_ID = matter_enums.ThermostatCluster.ID

  @property
  @abc.abstractmethod
  def local_temperature(self) -> int:
    """The LocalTemperature attribute.

    Returns:
      The temperature, as measured locally or remotely (over the network),
      on the Celsius scale with a resolution of 0.01°C.
    """

  @property
  @abc.abstractmethod
  def occupied_cooling_setpoint(self) -> int:
    """The OccupiedCoolingSetpoint attribute.

    Returns:
      The cooling mode setpoint when the room is occupied, on the Celsius scale
      with a resolution of 0.01°C.
    """

  @property
  @abc.abstractmethod
  def occupied_heating_setpoint(self) -> int:
    """The OccupiedHeatingSetpoint attribute.

    Returns:
      The heating mode setpoint when the room is occupied, on the Celsius scale
      with a resolution of 0.01°C.
    """

  @property
  @abc.abstractmethod
  def control_sequence_of_operation(
      self) -> matter_enums.ThermostatControlSequence:
    """The ControlSequenceOfOperation attribute.

    Returns:
      The overall operating environment of the thermostat, and the possible
      system modes that the thermostat can operate in.
    """

  @property
  @abc.abstractmethod
  def system_mode(self) -> matter_enums.ThermostatSystemMode:
    """The SystemMode attribute.

    Returns:
      The current operating mode of the thermostat.
    """

  @abc.abstractmethod
  def setpoint_raise_lower(self, mode: matter_enums.ThermostatSetpointMode,
                           amount: int) -> None:
    """The SetpointRaiseLower command to add specified amount to setpoint.

    Args:
      mode: The mode for which the sepoints are to be adjusted.
      amount: The amount (possibly negative) that should be added to the
        setpoint(s), in steps of 0.1°C
    """



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""chip-tool implementation of Matter Occupancy Sensing cluster capability.
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import occupancy_sensing_base

_CLUSTER_NAME = "occupancysensing"


class OccupancySensingClusterChipTool(
    occupancy_sensing_base.OccupancySensingClusterBase):
  """chip-tool based Matter Occupancy Sensing cluster capability."""

  @decorators.DynamicProperty
  def occupancy(self) -> int:
    """The Occupancy attribute.

    This attribute is a bitmap where bit 0 specifies the sensed occupancy as
    follows: 1 = occupied, 0 = unoccupied. All other bits are reserved.

    Returns:
      The Occupancy attribute.
    """
    return self._read(self._endpoint_id, _CLUSTER_NAME, "occupancy")

  @decorators.DynamicProperty
  def occupancy_sensor_type(self) -> matter_enums.OccupancySensorType:
    """The OccupancySensorType attribute.

    This attribute specifies the type of the occupancy sensor:
    0 = PIR; 1 = Ultrasonic; 2 = PIR and ultrasonic; 3 = Physical contact.

    Returns:
      The OccupancySensorType attribute.
    """
    sensor_type_data = self._read(self._endpoint_id, _CLUSTER_NAME,
                                  "occupancy-sensor-type")
    return matter_enums.OccupancySensorType(sensor_type_data)

  @decorators.DynamicProperty
  def occupancy_sensor_type_bitmap(self) -> int:
    """The OccupancySensorTypeBitmap attribute.

    The attribute specifies the types of the occupancy sensor: a "1" in each bit
    position indicates this type is implemented: Bit 0 = PIR; Bit 1 = Ultrasonic
    ; Bit 2 = Physical contact.

    Returns:
      The OccupancySensorTypeBitmap attribute.
    """
    return self._read(self._endpoint_id, _CLUSTER_NAME,
                      "occupancy-sensor-type-bitmap")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matter Heating Cooling Unit endpoint.

A Heating/Cooling Unit is a device capable of heating or cooling a space in a
house. It is not mandatory to provide both functionalities (for example, the
device may just heat but not cool). It may be an indoor air handler.

This endpoint module corresponds to the "Heating Cooling Unit" Matter device
type which defines required and optional clusters in the Matter spec 1.0.
"""
from typing import Union

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import fan_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import level_control_chip_tool
from gazoo_device.capabilities.matter_clusters import level_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import on_off_chip_tool
from gazoo_device.capabilities.matter_clusters import on_off_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import heating_cooling_unit_base


class HeatingCoolingUnitEndpoint(
    heating_cooling_unit_base.HeatingCoolingUnitBase):
  """Matter Heating Cooling Unit endpoint."""

  @decorators.CapabilityDecorator(
      [on_off_chip_tool.OnOffClusterChipTool, on_off_pw_rpc.OnOffClusterPwRpc])
  def on_off(
      self
  ) -> Union[on_off_chip_tool.OnOffClusterChipTool,
             on_off_pw_rpc.OnOffClusterPwRpc]:
    """Matter OnOff cluster instance."""
    return self.cluster_lazy_init(matter_enums.OnOffCluster.ID)

  @decorators.CapabilityDecorator([
      level_control_chip_tool.LevelControlClusterChipTool,
      level_control_pw_rpc.LevelControlClusterPwRpc
  ])
  def level_control(
      self
  ) -> Union[level_control_chip_tool.LevelControlClusterChipTool,
             level_control_pw_rpc.LevelControlClusterPwRpc]:
    """Matter Level Control cluster instance."""
    return self.cluster_lazy_init(matter_enums.LevelControlCluster.ID)

  @decorators.CapabilityDecorator(fan_control_pw_rpc.FanControlClusterPwRpc)
  def fan_control(self) -> fan_control_pw_rpc.FanControlClusterPwRpc:
    """Matter Fan Control cluster instance."""
    return self.cluster_lazy_init(matter_enums.FanControlCluster.ID)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matter Humidity Sensor endpoint.

A humidity sensor (in most cases a Relative humidity sensor) reports humidity
measurements.

This endpoint module corresponds to the
"Humidity Sensor" Matter device type which defines required and optional
clusters in the Matter spec 1.0.
A Humidity Sensor endpoint houses an instance of Humidity Sensor Matter device
type with the following defined clusters on this endpoint:

The required clusters for this endpoint: Relative Humidity Measurement
"""
from typing import Union

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import relative_humidity_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import relative_humidity_measurement_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import humidity_sensor_base


class HumiditySensorEndpoint(humidity_sensor_base.HumiditySensorBase):
  """Matter Humidity Sensor endpoint."""

  @decorators.CapabilityDecorator([
      relative_humidity_measurement_chip_tool
      .RelativeHumidityMeasurementClusterChipTool,
      relative_humidity_measurement_pw_rpc
      .RelativeHumidityMeasurementClusterPwRpc
  ])
  def relative_humidity_measurement(
      self
  ) -> Union[relative_humidity_measurement_chip_tool
             .RelativeHumidityMeasurementClusterChipTool,
             relative_humidity_measurement_pw_rpc
             .RelativeHumidityMeasurementClusterPwRpc]:
    """Matter Relative Humidity Measurement cluster instance."""
    return self.cluster_lazy_init(
        matter_enums.RelativeHumidityMeasurementCluster.ID)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mixin for Matter endpoint aliases."""
from gazoo_device import decorators
from gazoo_device.capabilities.matter_endpoints import color_temperature_light
from gazoo_device.capabilities.matter_endpoints import contact_sensor
from gazoo_device.capabilities.matter_endpoints import dimmable_light
from gazoo_device.capabilities.matter_endpoints import door_lock
from gazoo_device.capabilities.matter_endpoints import extended_color_light
from gazoo_device.capabilities.matter_endpoints import fan
from gazoo_device.capabilities.matter_endpoints import flow_sensor
from gazoo_device.capabilities.matter_endpoints import heating_cooling_unit
from gazoo_device.capabilities.matter_endpoints import humidity_sensor
from gazoo_device.capabilities.matter_endpoints import light_sensor
from gazoo_device.capabilities.matter_endpoints import occupancy_sensor
from gazoo_device.capabilities.matter_endpoints import on_off_light
from gazoo_device.capabilities.matter_endpoints import on_off_light_switch
from gazoo_device.capabilities.matter_endpoints import on_off_plugin_unit
from gazoo_device.capabilities.matter_endpoints import pressure_sensor
from gazoo_device.capabilities.matter_endpoints import root_node
from gazoo_device.capabilities.matter_endpoints import speaker
from gazoo_device.capabilities.matter_endpoints import temperature_sensor
from gazoo_device.capabilities.matter_endpoints import thermostat
from gazoo_device.capabilities.matter_endpoints import window_covering


class MatterEndpointAliasesMixin:
  """Mixin for Matter endpoint aliases.

  The mixin assumes self.matter_endpoints capability is set.
  """

  @decorators.CapabilityDecorator(
      color_temperature_light.ColorTemperatureLightEndpoint)
  def color_temperature_light(
      self) -> color_temperature_light.ColorTemperatureLightEndpoint:
    """Matter Color Temperature Light endpoint instance.

    Returns:
      Color Temperature Light endpoint instance.

    Raises:
      DeviceError when Color Temperate Light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        color_temperature_light.ColorTemperatureLightEndpoint)

  @decorators.CapabilityDecorator(contact_sensor.ContactSensorEndpoint)
  def contact_sensor(self) -> contact_sensor.ContactSensorEndpoint:
    """Matter Contact Sensor endpoint instance.

    Returns:
      Contact Sensor endpoint instance.

    Raises:
      DeviceError when Contact Sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        contact_sensor.ContactSensorEndpoint)

  @decorators.CapabilityDecorator(dimmable_light.DimmableLightEndpoint)
  def dimmable_light(self) -> dimmable_light.DimmableLightEndpoint:
    """Matter Dimmable Light endpoint instance.

    Returns:
      Dimmable Light endpoint instance.

    Raises:
      DeviceError when Dimmable Light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        dimmable_light.DimmableLightEndpoint)

  @decorators.CapabilityDecorator(door_lock.DoorLockEndpoint)
  def door_lock(self) -> door_lock.DoorLockEndpoint:
    """Matter Door Lock endpoint instance.

    Returns:
      Door Lock endpoint instance.

    Raises:
      DeviceError when Door Lock endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        door_lock.DoorLockEndpoint)

  @decorators.CapabilityDecorator(
      extended_color_light.ExtendedColorLightEndpoint)
  def extended_color_light(
      self) -> extended_color_light.ExtendedColorLightEndpoint:
    """Matter Extended color light endpoint instance.

    Returns:
      Extended color light endpoint instance.

    Raises:
      DeviceError when Extended color light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        extended_color_light.ExtendedColorLightEndpoint)

  @decorators.CapabilityDecorator(fan.FanEndpoint)
  def fan(self) -> fan.FanEndpoint:
    """Matter Fan endpoint instance.

    Returns:
      Fan endpoint instance.

    Raises:
      DeviceError when Fan endpoint is not supported on the device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(fan.FanEndpoint)

  @decorators.CapabilityDecorator(flow_sensor.FlowSensorEndpoint)
  def flow_sensor(self) -> flow_sensor.FlowSensorEndpoint:
    """Matter Flow sensor endpoint instance.

    Returns:
      Flow sensor endpoint instance.

    Raises:
      DeviceError when flow sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        flow_sensor.FlowSensorEndpoint)

  @decorators.CapabilityDecorator(
      heating_cooling_unit.HeatingCoolingUnitEndpoint)
  def heating_cooling_unit(
      self) -> heating_cooling_unit.HeatingCoolingUnitEndpoint:
    """Matter heating_cooling_unit endpoint instance.

    Returns:
      Heating cooling unit endpoint instance.

    Raises:
      DeviceError when Heating cooling unit endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        heating_cooling_unit.HeatingCoolingUnitEndpoint)

  @decorators.CapabilityDecorator(humidity_sensor.HumiditySensorEndpoint)
  def humidity_sensor(self) -> humidity_sensor.HumiditySensorEndpoint:
    """Matter Humidity Sensor endpoint instance.

    Returns:
      Humidity Sensor endpoint instance.

    Raises:
      DeviceError when Humidity Sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        humidity_sensor.HumiditySensorEndpoint)

  @decorators.CapabilityDecorator(light_sensor.LightSensorEndpoint)
  def light_sensor(self) -> light_sensor.LightSensorEndpoint:
    """Matter Light sensor endpoint instance.

    Returns:
      Light sensor endpoint instance.

    Raises:
      DeviceError when light sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        light_sensor.LightSensorEndpoint)

  @decorators.CapabilityDecorator(occupancy_sensor.OccupancySensorEndpoint)
  def occupancy_sensor(self) -> occupancy_sensor.OccupancySensorEndpoint:
    """Matter Occupancy Sensor endpoint instance.

    Returns:
      Occupancy Sensor endpoint instance.

    Raises:
      DeviceError when Occupancy Sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        occupancy_sensor.OccupancySensorEndpoint)

  @decorators.CapabilityDecorator(on_off_light.OnOffLightEndpoint)
  def on_off_light(self) -> on_off_light.OnOffLightEndpoint:
    """Matter OnOff Light endpoint instance.

    Returns:
      OnOff Light endpoint instance.

    Raises:
      DeviceError when OnOff Light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        on_off_light.OnOffLightEndpoint)

  @decorators.CapabilityDecorator(on_off_light_switch.OnOffLightSwitchEndpoint)
  def on_off_light_switch(self) -> on_off_light_switch.OnOffLightSwitchEndpoint:
    """Matter OnOff Light Switch endpoint instance.

    Returns:
      OnOff Light Switch endpoint instance.

    Raises:
      DeviceError when OnOff Light Switch endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        on_off_light_switch.OnOffLightSwitchEndpoint)

  @decorators.CapabilityDecorator(on_off_plugin_unit.OnOffPluginUnitEndpoint)
  def on_off_plugin_unit(self) -> on_off_plugin_unit.OnOffPluginUnitEndpoint:
    """Matter On/Off Plug-in Unit endpoint instance.

    Returns:
      On/Off Plug-in Unit endpoint instance.

    Raises:
      DeviceError when On/Off Plug-in Unit endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        on_off_plugin_unit.OnOffPluginUnitEndpoint)

  @decorators.CapabilityDecorator(pressure_sensor.PressureSensorEndpoint)
  def pressure_sensor(self) -> pressure_sensor.PressureSensorEndpoint:
    """Matter Pressure Sensor endpoint instance.

    Returns:
      Pressure Sensor endpoint instance.

    Raises:
      DeviceError when Pressure Sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        pressure_sensor.PressureSensorEndpoint)

  @decorators.CapabilityDecorator(root_node.RootNodeEndpoint)
  def root_node(self) -> root_node.RootNodeEndpoint:
    """Matter Root Node endpoint instance.

    Returns:
      Root Node endpoint instance.

    Raises:
      DeviceError when Root Node endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        root_node.RootNodeEndpoint)

  @decorators.CapabilityDecorator(speaker.SpeakerEndpoint)
  def speaker(self) -> speaker.SpeakerEndpoint:
    """Matter Speaker endpoint instance.

    Returns:
      Speaker endpoint instance.

    Raises:
      DeviceError when Speaker endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        speaker.SpeakerEndpoint)

  @decorators.CapabilityDecorator(temperature_sensor.TemperatureSensorEndpoint)
  def temperature_sensor(self) -> temperature_sensor.TemperatureSensorEndpoint:
    """Matter Temperature Sensor endpoint instance.

    Returns:
      Temperature Sensor endpoint instance.

    Raises:
      DeviceError when Temperature Sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        temperature_sensor.TemperatureSensorEndpoint)

  @decorators.CapabilityDecorator(thermostat.ThermostatEndpoint)
  def thermostat(self) -> thermostat.ThermostatEndpoint:
    """Matter Thermostat endpoint instance.

    Returns:
      Thermostat endpoint instance.

    Raises:
      DeviceError when Thermostat endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        thermostat.ThermostatEndpoint)

  @decorators.CapabilityDecorator(window_covering.WindowCoveringEndpoint)
  def window_covering(self) -> window_covering.WindowCoveringEndpoint:
    """Matter window covering endpoint instance.

    Returns:
      Window covering endpoint instance.

    Raises:
      DeviceError when Window Covering endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        window_covering.WindowCoveringEndpoint)
